'''
description: 
    this a script tool for cp2k input file constructing.
    The source data is if from github:https://github.com/deepmodeling/dpgen/blob/master/dpgen/generator/lib/cp2k.py

return {*}
author: wuxingxing
'''
import os
import numpy as np
default_config = {
    "GLOBAL": {"PROJECT": "AL_PWMLFF"},
    "FORCE_EVAL": {
        "METHOD": "QS",
        "STRESS_TENSOR": "ANALYTICAL",
        "DFT": {
            "BASIS_SET_FILE_NAME": "BASIS_MOLOPT",
            "POTENTIAL_FILE_NAME": "GTH_POTENTIALS",
            "CHARGE": 0,
            "UKS": "F",
            "MULTIPLICITY": 1,
            "MGRID": {"CUTOFF": 400, "REL_CUTOFF": 50, "NGRIDS": 4},
            "QS": {"EPS_DEFAULT": "1.0E-12"},
            "SCF": {"SCF_GUESS": "ATOMIC", "EPS_SCF": "1.0E-6", "MAX_SCF": 50},
            "XC": {"XC_FUNCTIONAL": {"_": "PBE"}},
        },
        "SUBSYS": {
            "CELL": {"A": "10 .0 .0", "B": ".0 10 .0", "C": ".0 .0 10"},
            "COORD": {"@include": "coord.xyz"},
            "KIND": {
                "_": ["H", "C", "N"],
                "POTENTIAL": ["GTH-PBE-q1", "GTH-PBE-q4", "GTH-PBE-q5"],
                "BASIS_SET": ["DZVP-MOLOPT-GTH", "DZVP-MOLOPT-GTH", "DZVP-MOLOPT-GTH"],
            },
        },
        "PRINT": {"FORCES": {"_": "ON"}, "STRESS_TENSOR": {"_": "ON"}},
    },
}


def update_dict(old_d, update_d):
    """A method to recursive update dict
    :old_d: old dictionary
    :update_d: some update value written in dictionary form.
    """
    import collections.abc

    for k, v in update_d.items():
        if (
            k in old_d
            and isinstance(old_d[k], dict)
            and isinstance(update_d[k], collections.abc.Mapping)
        ):
            update_dict(old_d[k], update_d[k])
        else:
            old_d[k] = update_d[k]


def iterdict(d, out_list, flag=None):
    """:doc: a recursive expansion of dictionary into cp2k input
    :k: current key
    :v: current value
    :d: current dictionary under expansion
    :flag: used to record dictionary state. if flag is None,
    it means we are in top level dict. flag is a string.
    """
    for k, v in d.items():
        k = str(k)  # cast key into string
        # if value is dictionary
        if isinstance(v, dict):
            # flag == None, it is now in top level section of cp2k
            if flag is None:
                out_list.append("&" + k)
                out_list.append("&END " + k)
                iterdict(v, out_list, k)
            # flag is not None, now it has name of section
            else:
                index = out_list.index("&END " + flag)
                out_list.insert(index, "&" + k)
                out_list.insert(index + 1, "&END " + k)
                iterdict(v, out_list, k)
        elif isinstance(v, list):
            #            print("we have encountered the repeat section!")
            index = out_list.index("&" + flag)
            # delete the current constructed repeat section
            del out_list[index : index + 2]
            # do a loop over key and corresponding list
            k_tmp_list = []
            v_list_tmp_list = []
            for k_tmp, v_tmp in d.items():
                k_tmp_list.append(str(k_tmp))
                v_list_tmp_list.append(v_tmp)
            for repeat_keyword in zip(*v_list_tmp_list):
                out_list.insert(index, "&" + flag)
                out_list.insert(index + 1, "&END " + flag)
                for idx, k_tmp in enumerate(k_tmp_list):
                    if k_tmp == "_":
                        out_list[index] = "&" + flag + " " + repeat_keyword[idx]
                    else:
                        out_list.insert(index + 1, k_tmp + " " + repeat_keyword[idx])

            break

        else:
            v = str(v)
            if flag is None:
                out_list.append(k + " " + v)
                print(k, ":", v)
            else:
                if k == "_":
                    index = out_list.index("&" + flag)
                    out_list[index] = "&" + flag + " " + v

                else:
                    index = out_list.index("&END " + flag)
                    out_list.insert(index, k + " " + v)


def make_cp2k_input(
        cell:list[float],
        atom_names:list[str], 
        basis_set_file_name:str,
        potential_file_name:str,
        xc_functional:str,
        potential:dict,
        basis_set:dict,
        coord_content:str = None # not used
        ):
    # covert cell to cell string
    cell = np.reshape(cell, [3, 3])
    cell_a = np.array2string(cell[0, :])
    cell_a = cell_a[1:-1]
    cell_b = np.array2string(cell[1, :])
    cell_b = cell_b[1:-1]
    cell_c = np.array2string(cell[2, :])
    cell_c = cell_c[1:-1]
    # get update from cell
    in_potentail =[]
    in_basis_set = []
    for atom in atom_names:
        in_potentail.append(potential[atom])
        in_basis_set.append(basis_set[atom])

    cell_config = {
        "FORCE_EVAL": {
            "SUBSYS": {
                        "BASIS_SET_FILE_NAME": basis_set_file_name,
                        "POTENTIAL_FILE_NAME": potential_file_name,
                        "CELL": {"A": cell_a, "B": cell_b, "C": cell_c},
                        # "COORD": coord_content, 
                        "COORD": {"@include": "coord.xyz"}, 
                        "KIND": {
                        "_": atom_names,
                        "POTENTIAL": in_potentail,
                        "BASIS_SET": in_basis_set
                        },
                        "XC": {"XC_FUNCTIONAL": {"_": xc_functional}}
            }
        }
    }
    update_dict(default_config, cell_config)
    # output list
    input_str = []
    iterdict(default_config, input_str)
    string = "\n".join(input_str)
    return string

def make_cp2k_xyz(atom_types:list[int], coord_list:list[float]):
    # get structral information
    # write coordinate to xyz file used by cp2k input
    atom_list = np.array(atom_types)
    x = "\n"
    for kind, coord in zip(atom_list, coord_list):
        x += str(kind) + " " + str(coord[:])[1:-1] + "\n"
    return x

'''
description: 
    delete cell and coord
    1.for coord, set it :"COORD": {"@include": "coord.xyz"}
    2.for cell, set it with input
    3.set 'PRINT_LEVEL' to 'medium'
    attation, the cell block and coord block has no order
param {*} sys_data
param {*} exinput_path
return {*}
author: wuxingxing
'''
def make_cp2k_input_from_external(cell, coord_file, exinput_path, gaussian_base_param:dict):
    coord_file_name = os.path.basename(coord_file)
    # insert the cell information
    # covert cell to cell string
    cell = np.reshape(cell, [3, 3])
    # read the input content as string
    with open(exinput_path) as f:
        exinput = f.readlines()
    exinput = [line.upper() for line in exinput]
    # replace the cell string
    start_cell = 0
    end_cell = 0
    start_subsys = 0
    end_subsys = 0
    start_coord = 0
    end_coord = 0
    start_kind = -1
    end_kind = -1
    start_global = 0
    end_global = 0
    print_level_line = -1
    start_dft = 0
    end_dft = 0
    basis_set_file_name = -1
    potential_file_name = -1
    start_kpoint = -1
    end_kpoint = -1
    kpoint_line = 0
    # delete the BASIS_SET_FILE_NAME and POTENTIAL_FILE_NAME line
    for line_idx, line in enumerate(exinput):
        line = line.upper()
        if "&DFT" in line:
            start_dft = line_idx
        if "&END DFT" in line:
            end_dft = line_idx
        if "BASIS_SET_FILE_NAME" in line:
            basis_set_file_name = line_idx
        if "POTENTIAL_FILE_NAME" in line:
            potential_file_name = line_idx
        if "&KPOINTS" in line:
            start_kpoint = line_idx
        if "&END KPOINTS" in line:
            end_kpoint = line_idx
        if "MONKHORST-PACK" in line:
            kpoint_line = line_idx
    
    if start_dft == end_dft:
        raise Exception("{} extarcted error! Can not find DFT set!".format(exinput_path))
    
    # set kspacing and pseudo file
    if gaussian_base_param["KSPACING"] is not None and kpoint_line > 0:
        raise Exception("The 'kspacing' in 'gaussian_param' and 'KPOINTS' in {} file cannot be set set simultaneously!".format(os.path.basename(exinput_path)))
    elif gaussian_base_param["KSPACING"] is None and kpoint_line == 0:
        kspacing_content = make_kspacing_kpoints(cell, 0.5)
    elif gaussian_base_param["KSPACING"] is not None:
        kspacing_content = make_kspacing_kpoints(cell, gaussian_base_param["KSPACING"])
    else:
        kspacing_content = "\n".join(exinput[start_kpoint:end_kpoint+1])
    basis_set_file_name, potential_file_name = sorted([basis_set_file_name, potential_file_name], reverse=True)
    if kpoint_line > 0:
        if start_kpoint > basis_set_file_name:
            for del_idx in list(range(end_kpoint, start_kpoint-1, -1)):
                exinput.pop(del_idx)
            if basis_set_file_name != -1:
                exinput.pop(basis_set_file_name)
            if potential_file_name != -1:
                exinput.pop(potential_file_name)
        else:
            if basis_set_file_name != -1:
                exinput.pop(basis_set_file_name)
            if potential_file_name != -1:
                exinput.pop(potential_file_name)
            for del_idx in list(range(end_kpoint, start_kpoint-1, -1)):
                exinput.pop(del_idx)
    else:
        if basis_set_file_name != -1:
            exinput.pop(basis_set_file_name)
        if potential_file_name != -1:
            exinput.pop(potential_file_name)
    
    exinput.insert(start_dft+1, "    BASIS_SET_FILE_NAME {}\n".format(gaussian_base_param["BASIS_SET_FILE_NAME"]))
    exinput.insert(start_dft+2, "    POTENTIAL_FILE_NAME {}\n".format(gaussian_base_param["POTENTIAL_FILE_NAME"]))
    exinput.insert(start_dft+3, kspacing_content)
    
    for line_idx, line in enumerate(exinput):
        line = line.upper()
        if "&GLOBAL" in line:
            start_global = line_idx
        if "&END GLOBAL" in line:
            end_global = line_idx
        if "PRINT_LEVEL" in line:
            print_level_line = line_idx
        if "&SUBSYS" in line:
            start_subsys = line_idx
        if "&END SUBSYS" in line:
            end_subsys = line_idx
        if "&CELL" in line:
            start_cell = line_idx
        if "&END CELL" in line:
            end_cell = line_idx
        if "&COORD" in line:
            start_coord = line_idx
        if "&END COORD" in line:
            end_coord = line_idx
        if "&KIND" in line and start_kind == -1:
            start_kind = line_idx
        if "&END KIND" in line:
            end_kind = line_idx

    if start_global == end_global:
        raise Exception("ERROR! the input cp2k inp file does not have 'GLOBAL' block! Please check the file {}\n".format(exinput_path))
    
    temp_exinput = exinput[:start_subsys+1]
    # add coord
    temp_exinput.append("    &COORD\n")
    temp_exinput.append("        @include {}\n".format(coord_file_name))
    # temp_exinput.append("        COORD_FILE_FORMAT XYZ\n")
    temp_exinput.append("    &END COORD\n")
    #add cell
    temp_exinput.append("    &CELL\n")
    temp_exinput.append("        A     {}    {}     {}\n".format(cell[0, 0], cell[0, 1], cell[0, 2]))
    temp_exinput.append("        B     {}    {}     {}\n".format(cell[1, 0], cell[1, 1], cell[1, 2]))
    temp_exinput.append("        C     {}    {}     {}\n".format(cell[2, 0], cell[2, 1], cell[2, 2]))
    # temp_exinput.append("        PERIODIC XYZ\n")
    temp_exinput.append("    &END CELL\n")

    kind_input = get_kind(coord_file=coord_file, gassion_base_param=gaussian_base_param)
    temp_exinput.append(kind_input)

    del_content_index = []
    if start_cell != end_cell:
        del_content_index.extend(list(range(start_cell, end_cell+1)))
    if start_coord != end_coord:
        del_content_index.extend(list(range(start_coord, end_coord+1)))
    if start_kind != end_kind:
        del_content_index.extend(list(range(start_kind, end_kind+1)))
    del_content_index = sorted(del_content_index)
    for index in range(start_subsys+1, end_subsys):
        if index not in del_content_index:
            temp_exinput.append(exinput[index])
    temp_exinput.extend(exinput[end_subsys:])

    #reset or add print_level medium
    if print_level_line == -1:
        temp_exinput.insert(end_global-1, "    PRINT_LEVEL medium\n")
    else:
        temp_exinput[print_level_line] = "    PRINT_LEVEL medium\n"
    return "".join(temp_exinput)

def get_kind(coord_file:str, gassion_base_param:dict):
    atom_list = get_atom_type_from_config(coord_file)
    kind_line = "\n"
    for idx, atom in enumerate(gassion_base_param["ELEMENT"]):
        if atom not in atom_list:
            continue
        kind_line += "    &KIND {}\n".format(atom)
        kind_line += "        ELEMENT {}\n".format(atom)
        kind_line += "        BASIS_SET {}\n".format(gassion_base_param["BASIS_SET"][idx])
        kind_line += "        POTENTIAL {}\n".format(gassion_base_param["POTENTIAL"][idx])
        kind_line += "    &END KIND\n"
    return kind_line

def get_atom_type_from_config(coord_file:str):
    res = []
    with open(coord_file, 'r') as rf:
        lines = rf.readlines()
    for line in lines:
        try:
            atom_type, x, y, z = line.strip().split()
            x = float(x)
            y = float(y)
            z = float(z)
            if atom_type not in res:
                res.append(atom_type)
        except:
            continue
    return res

def _reciprocal_box(box):
    rbox = np.linalg.inv(box)
    rbox = rbox.T
    return rbox

def make_kspacing_kpoints(lattice, kspacing):
    rbox = _reciprocal_box(lattice)
    kpoints = [
        round(2 * np.pi * np.linalg.norm(ii) / kspacing) for ii in rbox
    ]
    kpoints[0] = 1 if kpoints[0] == 0 else kpoints[0]
    kpoints[1] = 1 if kpoints[1] == 0 else kpoints[1]
    kpoints[2] = 1 if kpoints[2] == 0 else kpoints[2]
    ret = "    &KPOINTS\n"
    ret += "        SCHEME MONKHORST-PACK %d %d %d\n" % (kpoints[0], kpoints[1], kpoints[2])
    ret += "    &END KPOINTS\n"
    return ret
    # ret = _make_pwmat_kp_mp(kpoints)

# if __name__=="__main__":
#     import dpdata
#     poscar = "/data/home/wuxingxing/datas/al_dir/si_4_vasp/init_bulk/collection/init_config_0/0.9_scale.poscar"
#     sys_data = dpdata.System(poscar).data
    
#     from pwdata.config import Configs
#     from pwdata.calculators.const import ELEMENTTABLE_2
#     image = Configs.read(format="pwmat", data_path="/data/home/wuxingxing/datas/al_dir/si_exp/init_bulk/atom.config")
#     image = image._set_cartesian() if image.cartesian is False else image._set_cartesian()
#     potential = {"Si":"GTH-PBE"}
#     basis_set = {"Si":"DZVP-MOLOPT-SR-GTH-q4"}
#     atom_types_image = []
#     for atom in image.atom_types_image:
#         atom_types_image.append(ELEMENTTABLE_2[atom])
#     coord_xyz = make_cp2k_xyz(
#         atom_types = atom_types_image,
#         coord_list = image.position
#     )
#     with open("/data/home/wuxingxing/datas/al_dir/si_exp/init_bulk/coord.xyz", "w") as fp:
#         fp.write(coord_xyz)
    
#     make_cp2k_input(
#         cell = image.lattice,
#         atom_names=["Si"],
#         basis_set_file_name="BASIS_SET_FILE",
#         potential_file_name="POTENTIAL_FILE",
#         xc_functional="PBE",
#         potential=potential,
#         basis_set=basis_set,
#         coord_content=coord_xyz
#     )


import numpy as np
from pwact.utils.json_operation import get_parameter, get_required_parameter

class Descriptor(object):
    def __init__(self, json_input:dict, model_type:str, cmd:str, feature_type:list=None, type_embedding:bool=False) -> None:
        self.model_type = model_type
        self.cmd = cmd
        self.feature_type = feature_type
        self.supported_feature_group = [[1, 2], [3, 4], [5], [6], [7], [8]]
        self.Rmax = get_parameter("Rmax", json_input, 6.0)
        self.Rmin = get_parameter("Rmin", json_input, 0.5)
        self.M2 = get_parameter("M2", json_input, 16)
        self.E_tolerance = get_parameter("E_tolerance", json_input, 9999999.0)
        self.cheby_order = get_parameter("cheby_order", json_input, 10)
        self.radial_num1 = get_parameter("radial_num1", json_input, 25)
        self.radial_num2 = get_parameter("radial_num2", json_input, 4)
        if self.feature_type is not None:   #if feature type specified at first layer of json 
            if "feature_type" in json_input.keys(): # if feature type sepcified at descriptor layer of json
                if self.feature_type != json_input["feature_type"]: # if this two values are different, show WARNING info.
                    print("\n\nWARNNING: feature_type is specified twice with inconsistent values. \nThe program will prioritize the value specified within the Descriptor.\n\n")
                    self.feature_type = json_input["feature_type"]
        else:   # the first layer not specified in both first layer and descriptor layer, use default value
            self.feature_type = [int(_) for _ in get_parameter("feature_type", json_input, [3,4])]

        # dp params These parameters are only used for outputting to the standard output json file
        self.network_size = get_parameter("network_size", json_input, [25, 25, 25])
        self.bias = get_parameter("bias", json_input, True)
        self.resnet_dt = get_parameter("resnet_dt", json_input, False) # resnet in embedding net is False.
        self.activation = get_parameter("activation", json_input, "tanh")

        # type embedding params These parameters are only used for outputting to the standard output json file
        type_embedding_dict = get_parameter("type_embedding", json_input, {})
        self.type_embedding = False if (not type_embedding_dict.items()) and type_embedding is False else True
        # set dp embedding net params
        self.type_physical_property = get_parameter("physical_property", type_embedding_dict, ["atomic_number", "atom_radius", "atom_mass", "electron_affin", "pauling"])
        self.type_network_size = get_parameter("network_size", type_embedding_dict, None)
        if self.type_network_size is not None:
            self.type_network_size = [len(self.type_physical_property)] + self.type_network_size

        self.type_bias = get_parameter("bias", type_embedding_dict, True)
        self.type_resnet_dt = get_parameter("resnet_dt", type_embedding_dict, False) # resnet in embedding net is False.
        self.type_activation = get_parameter("activation", type_embedding_dict, "tanh")

        if self.feature_type not in self.supported_feature_group:
            raise Exception("ERROR: The input feature type group {} is not support, \
                            we currently support the following combinations:\n {}".format(self.feature_type, self.supported_feature_group))
        self.feature_dict = {}  # for feature generating
        self.feature_dict_out = {}  # for std output json
        for ftype in self.feature_type:
            ftype = "{}".format(ftype)
            ftype_dict = get_parameter("{}".format(ftype), json_input, {})
            # self.feature_dict[ftype] = getattr(self, "set_ftype{}_para".format(ftype), ftype_dict)
            if '1' == ftype:
                self.feature_dict[ftype], self.feature_dict_out[ftype] = self.set_ftype1_para(ftype_dict)
            elif '2' == ftype:
                self.feature_dict[ftype], self.feature_dict_out[ftype] = self.set_ftype2_para(ftype_dict)
            elif '3' == ftype:
                self.feature_dict[ftype], self.feature_dict_out[ftype] = self.set_ftype3_para(ftype_dict)
            elif '4' == ftype:
                self.feature_dict[ftype], self.feature_dict_out[ftype] = self.set_ftype4_para(ftype_dict)
            elif '5' == ftype:
                self.feature_dict[ftype], self.feature_dict_out[ftype] = self.set_ftype5_para(ftype_dict)
            elif '6' == ftype:
                self.feature_dict[ftype], self.feature_dict_out[ftype] = self.set_ftype6_para(ftype_dict)
            elif '7' == ftype:
                self.feature_dict[ftype], self.feature_dict_out[ftype] = self.set_ftype7_para(ftype_dict)
            elif '8' == ftype:
                self.feature_dict[ftype], self.feature_dict_out[ftype] = self.set_ftype8_para(ftype_dict)
        # for nep param
        self.cutoff = None
        self.n_max = None
        self.basis_size = None
        self.l_max = None
        self.type_weight = None
    '''
    description: 
    the return dict is used to assignment for value Ftype1_para in dfault_param.py
        the 'n2b' from json file is 'numOf2bfeat' value in Ftype1_para
    param {*} self
    param {dict} ftype_dict
    return {*}
    author: wuxingxing
    '''
    def set_ftype1_para(self, ftype_dict:dict):
        numOf2bfeat = get_parameter("n2b", ftype_dict, 24)
        iflag_grid = get_parameter("iflag_grid", ftype_dict, 3)
        fact_base = get_parameter("fact_base", ftype_dict, 0.2)
        dR1 = get_parameter("dR1", ftype_dict, 0.5)
        iflag_ftype = get_parameter("iflag_ftype", ftype_dict, 3)
        return {
            "numOf2bfeat":[numOf2bfeat for tmp in range(10)],
            "Rc": [self.Rmax for tmp in range(10)],
            "Rm": [self.Rmin for tmp in range(10)],
            "iflag_grid": [iflag_grid for tmp in range(10)],   # 1 or 2 or 3
            "fact_base": [fact_base for tmp in range(10)],
            "dR1": [dR1 for tmp in range(10)],
            "iflag_ftype": iflag_ftype  
        },  \
        {
            "n2b":numOf2bfeat,
            "iflag_grid": iflag_grid,   # 1 or 2 or 3
            "fact_base": fact_base,
            "dR1": dR1,
            "iflag_ftype": iflag_ftype  
        } 


    '''
    description: 
    the return dict is used to assignment for value Ftype2_para in dfault_param.py
        the 'n3b1' from json file is 'numOf3bfeat1' value in Ftype2_para
        the 'n3b2' from json file is 'numOf3bfeat2' value in Ftype2_para
    param {*} self
    param {dict} ftype_dict
    return {*}
    author: wuxingxing
    '''

    def set_ftype2_para(self, ftype_dict:dict):
        numOf3bfeat1 = get_parameter("n3b1", ftype_dict, 3)
        numOf3bfeat2 = get_parameter("n3b2", ftype_dict, 3)
        iflag_grid = get_parameter("iflag_grid", ftype_dict, 3)
        fact_base = get_parameter("fact_base", ftype_dict, 0.2)
        dR1 = get_parameter("dR1", ftype_dict, 0.5)
        dR2 = get_parameter("dR2", ftype_dict, 0.5)
        iflag_ftype = get_parameter("iflag_ftype", ftype_dict, 3)

        return {
            "numOf3bfeat1":[numOf3bfeat1 for tmp in range(10)],
            "numOf3bfeat2":[numOf3bfeat2 for tmp in range(10)],
            "Rc":[self.Rmax for tmp in range(10)],
            "Rc2":[self.Rmax for tmp in range(10)],
            "Rm":[self.Rmin for tmp in range(10)],
            "iflag_grid":[iflag_grid for tmp in range(10)],    # 1 or 2 or 3
            "fact_base":[fact_base for tmp in range(10)],
            "dR1":[dR1 for tmp in range(10)],
            "dR2":[dR2 for tmp in range(10)],
            "iflag_ftype":iflag_ftype   # same value for different types, iflag_ftype:1,2,3 when 3, iflag_grid must be 3
        }, \
         {
            "n3b1":numOf3bfeat1,
            "n3b2":numOf3bfeat1,
            "dR1":dR1,
            "dR2":dR2,
            "iflag_grid":iflag_grid,    # 1 or 2 or 3
            "fact_base":fact_base,
            "iflag_ftype":iflag_ftype   # same value for different types, iflag_ftype:1,2,3 when 3, iflag_grid must be 3
        }

    def set_ftype3_para(self, ftype_dict:dict):
        n2b = get_parameter("n2b", ftype_dict, 6)
        w = get_parameter("w", ftype_dict, [1.0, 1.5, 2.0])

        return {
            "Rc":[self.Rmax for tmp in range(10)],     # 5.4 number of elements in Rc = num atom type
            "n2b":[n2b for tmp in range(10)],       # 6 number of elements in n2b = num atom type
            "w": w #[1.0, 1.5, 2.0],
        }, \
        {
            "n2b":n2b,
            "w": w
        }
        # 1/w^2 is the \eta in formula, and w is the width of gaussian fuction

    def set_ftype4_para(self, ftype_dict:dict):
        n3b = get_parameter("n3b", ftype_dict, 20)
        zeta = get_parameter("zeta", ftype_dict, 2.0)
        w = get_parameter("w", ftype_dict, [0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0])
        return {
            "Rc":[self.Rmax for tmp in range(10)],     # 5.4 number of elements in Rc = num atom type
            "n3b":[n3b for tmp in range(10)],   # 20
            # n3b must be less than n_zeta * n_w * n_lambda, by default, n3b < 7*10*2=126
            # it is better if n3b is divisible by (n_w*2)
            "zeta": [ (zeta ** np.array(range(20))).tolist() for tmp in range(10)],  #2.0 feature changed
            "w":    [ w for tmp in range(10)],  # [0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0] feature changed
            # 1/w^2 is the \eta in formula, and w is the width of gaussian fuction
            # 'lambda':[ [1.0,-1.0] for tmp in range(10)], # lambda === [1.0, -1.0]
        },\
        {
            "n3b":n3b, 
            "zeta": zeta, 
            "w": w 
        }
        # 1/w^2 is the \eta in formula, and w is the width of gaussian fuction

    def set_ftype5_para(self, ftype_dict:dict):
        n_MTP_line = get_parameter("n_MTP_line", ftype_dict, 5)
        return {
            "Rc":[self.Rmax for tmp in range(10)],     # number of elements in Rc = num atom type
            "Rm":[self.Rmin  for tmp in range(10)],     # number of elements in Rc = num atom type
            "n_MTP_line": [n_MTP_line for tmp in range(10)], # 5~14
            
            "tensors":[
                        [
                        '1, 4, 0, ( )                              ',
                        '2, 3,3, 0,0, ( ), ( )                     ',
                        '2, 3,3, 1,1, ( 21 ), ( 11 )               ',
                        '2, 3,3, 2,2, ( 21, 22 ), ( 11, 12 )       ',
                        '3, 2,2,2, 2,1,1 ( 21, 31 ), ( 11 ), ( 12 )',
                        '3, 3,3,3, 2,1,1 ( 21, 31 ), ( 11 ), ( 12 )',
                        '3, 2,2,2, 3,2,1 ( 21, 22, 31 ), ( 11, 12 ), ( 13 )',
                        '3, 3,3,3, 3,2,1 ( 21, 22, 31 ), ( 11, 12 ), ( 13 )',
                        '3, 2,2,2, 4,2,2 ( 21, 22, 31, 32 ), ( 11, 12 ), ( 13, 14 )',
                        '3, 3,3,3, 4,2,2 ( 21, 22, 31, 32 ), ( 11, 12 ), ( 13, 14 )',
                        '4, 2,2,2,2 3,1,1,1 ( 21, 31, 41 ), ( 11 ), ( 12 ), ( 13 )',
                        '4, 3,3,3,3 3,1,1,1 ( 21, 31, 41 ), ( 11 ), ( 12 ), ( 13 )',
                        '4, 2,2,2,2 4,2,1,1 ( 21, 22, 31, 41 ), ( 11, 12 ), ( 13 ), ( 14 )',
                        '4, 3,3,3,3 4,2,1,1 ( 21, 22, 31, 41 ), ( 11, 12 ), ( 13 ), ( 14 )',
                        ] for tmp in range(10)
                    ],
            }, \
            {
            "n_MTP_line": n_MTP_line # 5~14
            }

    def set_ftype6_para(self, ftype_dict:dict):
        J = get_parameter("J", ftype_dict, 3.0)
        n_w_line = get_parameter("n_w_line", ftype_dict, 2)
        w1 = get_parameter("w1", ftype_dict, [0.9, 0.1, 0.8, 0.2, 0.7, 0.3, 0.6, 0.4])
        w2 = get_parameter("w2", ftype_dict, [0.1, 0.9, 0.2, 0.8, 0.3, 0.7, 0.3, 0.6])
        
        return {
            'Rc':[self.Rmax for tmp in range(10)],     # number of elements in Rc = num atom type
            'J' :[J for tmp in range(10)],  #3.0
            'n_w_line': [n_w_line for tmp in range(10)],#2
            'w1':[ w1 for tmp in range(10)],  # shape(w1) = (ntype, n_w_line) [0.9, 0.1, 0.8, 0.2, 0.7, 0.3, 0.6, 0.4]
            'w2':[ w2 for tmp in range(10) ], # [0.1, 0.9, 0.2, 0.8, 0.3, 0.7, 0.3, 0.6] 
        }, \
            {
            'J' :J,
            'n_w_line': n_w_line,
            'w1':w1,
            'w2':w2
        }

    def set_ftype7_para(self, ftype_dict:dict):
        M = get_parameter("M", ftype_dict, 25)
        M2 = get_parameter("M2", ftype_dict, 4)
        weight_r = get_parameter("weight_r", ftype_dict, 1.0)
        
        return {
            'Rc':[self.Rmax  for tmp in range(10)],     # number of elements in Rc = num atom type
            'Rc2':[self.Rmin  for tmp in range(10)],
            'Rm':[1.0 for tmp in range(10)],    #Chebyshev param , not rmin
            'M': [M for tmp in range(10)],#25
            'M2': [M2  for tmp in range(10)],#4
            'weight_r': [weight_r  for tmp in range(10)], #1.0
        }, \
         {
            'M': M,#25
            'M2': M2,#4
            'weight_r': weight_r #1.0
        }

    def set_ftype8_para(self, ftype_dict:dict):
        M = get_parameter("M", ftype_dict, 8)
        weight_r = get_parameter("weight_r", ftype_dict, 1.0)
        w = get_parameter("w", ftype_dict, [1.0, 1.5, 2.0, 2.5 ] )
        return {
            'Rc':[self.Rmax for tmp in range(10)],     # number of elements in Rc = num atom type
            'M':[M for tmp in range(10)], # 8
            'weight_r':[weight_r for tmp in range(10)], #1.0
            'w':w
        }, \
        {
            'M':M, # 8
            'weight_r':weight_r, #1.0
            'w':w
        }
    
    def to_dict(self):
        dicts = {}

        # dicts["E_tolerance"] = self.E_tolerance
        if self.model_type == "DP".upper():
            dicts["Rmax"] = self.Rmax
            dicts["Rmin"] = self.Rmin
            dicts["M2"] = self.M2
            dicts["network_size"] = self.network_size
            dicts["M2"] = self.M2
            # dicts["bias"] = self.bias
            # dicts["resnet_dt"] = self. resnet_dt 
            # dicts["activation"] = self.activation
        elif self.model_type == "NN".upper() or self.model_type == "Linear".upper():
            dicts["Rmax"] = self.Rmax
            dicts["Rmin"] = self.Rmin
            dicts["feature_type"] = self.feature_type
            for feature in self.feature_type:
                feature = "{}".format(feature)
                dicts[feature] = self.feature_dict_out[feature]
        elif self.model_type == "CHEBY".upper():
            dicts["cheby_order"] = self.cheby_order
            dicts["radial_num1"] = self.radial_num1
            dicts["radial_num2"] = self.radial_num2
        elif self.model_type.upper() == "NEP":
            dicts["cutoff"] = self.cutoff
            if self.zbl is not None:
                dicts["zbl"] = self.zbl
            dicts["n_max"] = self.n_max
            dicts["basis_size"] = self.basis_size
            dicts["l_max"] = self.l_max
            # dicts["type_weight"] = self.type_weight
        else:
            raise Exception("descriptor to dict: the model type not realized:{}".format(self.model_type))
        
        if self.type_embedding is True:
            dicts["type_embedding"] = {}
            dicts["type_embedding"]["physical_property"] = self.type_physical_property
            if self.type_network_size is not None:
               dicts["type_embedding"]["network_size"] = self.type_network_size[1:]
           
        return dicts


import glob
import re
import os
from math import ceil
from pwact.utils.constant import DFT_STYLE
GPU_SCRIPT_HEAD = \
"#!/bin/sh\n\
#SBATCH --job-name={}\n\
#SBATCH --nodes={}\n\
#SBATCH --ntasks-per-node={}\n\
#SBATCH --gres=gpu:{}\n\
#SBATCH --partition={}\n\
\
"

CPU_SCRIPT_HEAD = \
"#!/bin/sh\n\
#SBATCH --job-name={}\n\
#SBATCH --nodes={}\n\
#SBATCH --ntasks-per-node={}\n\
#SBATCH --partition={}\n\
\
"

'''
Description: 
Obtain the execution status of the slurm jobs under the 'dir'
    the slurm.job and tag format are number-jobname.job, number-tag-type-success.
    for example:
        iter.0000/explore/md/
        0-md.job 1-md.job 2-md.job 3-md.job 4-md.job 5-md.job 
        0-tag.md.success 1-tag.md.success 2-tag.md.success 3-tag.md.success 4-tag.md.success 5-tag.md.success

        or

        0-scf.job 1-scf.job 2-scf.job 3-scf.job 4-scf.job
        0-tag.scf.success 1-tag.scf.success 2-tag.scf.success 3-tag.scf.success 4-tag.scf.success
    new change:
        only jugt the slurm file is done by the tag under each subwork tag. so the tag of job file is nouse maybe
param {*} dir
Returns: 
Author: WU Xingxing
'''
def get_slurm_job_run_info(dir:str, job_patten:str="*.job", tag_patten:str="tag.*.success", for_back:bool=False):
    slurm_job_files = sorted(glob.glob(os.path.join(dir, job_patten)))
    slrum_job_dirs = [int(os.path.basename(_).split('-')[0]) for _ in slurm_job_files]

    slurm_tag_sucess_files = sorted(glob.glob(os.path.join(dir, tag_patten)))
    slrum_tag_sucess_dirs = [int(os.path.basename(_).split('-')[0]) for _ in slurm_tag_sucess_files]
    
    slurm_failed = []
    slurm_success = []

    if for_back is False:
        for slurm_file in slurm_job_files:
            if slurm_job_done_by_jobtag(slurm_file):
                slurm_success.append(slurm_file)
            else:
                slurm_failed.append(slurm_file)
    else:
        for i, d in enumerate(slrum_job_dirs):
            if d in slrum_tag_sucess_dirs:
                slurm_success.append(slurm_job_files[i])
            else:
                slurm_failed.append(slurm_job_files[i])
            
    return slurm_failed, slurm_success


def slurm_job_done_by_jobtag(slurm_file):
    with open(slurm_file, 'r') as f:
        content = f.read()
    cd_pattern = r'cd\s+([^\n]+)'
    directories = re.findall(cd_pattern, content)        
    if not directories:
        raise Exception("Error! There is no task in the slurm.job file {}".format(slurm_file))
    for directory in directories:
        directory = directory.strip()
        success_file = glob.glob(os.path.join(directory, "*.success"))
        if len(success_file) > 0:
            continue
        else:
            return False
    return True


def recheck_slurm_by_jobtag(slurm_files:list[str], tag):
    remain_job = []
    for slurm_file in slurm_files:
        with open(slurm_file, 'r') as f:
            content = f.read()
        cd_pattern = r'cd\s+([^\n]+)'
        directories = re.findall(cd_pattern, content)        
        if not directories:
            raise Exception("Error! There is no task in the slurm.job file {}".format(slurm_file))
        for directory in directories:
            directory = directory.strip()
            success_file = os.path.join(directory, tag)
            if os.path.exists(success_file):
                continue
            else:
                remain_job.append(slurm_file)
                break
    return remain_job

# def slurm_job_is_done_by_jobtag(dir:str, job_patten:str="*.job", tag_patten:str="tag.*.success"):
#     slurm_job_files = sorted(glob.glob(os.path.join(dir, job_patten)))
#     slurm_failed = []
#     for slurm_file in slurm_job_files:
#         with open(slurm_file, 'r') as f:
#             content = f.read()
#         cd_pattern = r'cd\s+([^\n]+)'
#         directories = re.findall(cd_pattern, content)        
#         if not directories:
#             raise Exception("Error! There is no task in the slurm.job file {}".format(slurm_file))
#         for directory in directories:
#             directory = directory.strip()
#             success_file = os.path.join(directory, tag_patten)
#             if os.path.exists(success_file):
#                 continue
#             else:
#                 slurm_failed.append(slurm_file)
#                 break
#     return slurm_failed


'''
description: 
    split the job_list with groupsize
param {int} groupsize
param {list} job_list: N jobs
return {*} [["job1","job2",...,"job_groupseze"], ..., [..., "job_N", "NONE",...,"NONE"]]
author: wuxingxing
'''
def split_job_for_group(groupsize:int , job_list:list[str], parallel_num=1):
    groupsize = len(job_list) if groupsize == -1 else groupsize
    if groupsize > 1:
        groupsize_adj = ceil(groupsize/parallel_num)
        if groupsize_adj*parallel_num > groupsize:
            groupsize_adj = ceil(groupsize/parallel_num)*parallel_num
            print("the groupsize automatically adjusts from {} to {}".format(groupsize, groupsize_adj))
            groupsize = groupsize_adj
            
    group_num = len(job_list) // groupsize
    sub_list = []
    for i in range(group_num):
        sub_list.append(job_list[i*groupsize:(i+1)*groupsize])
    
    last_sub = []
    if len(sub_list)*groupsize < len(job_list):
        for sub_index in range(len(sub_list)*groupsize, len(job_list)):
            last_sub.append(job_list[sub_index])
        while len(last_sub) < groupsize:
            last_sub.append("NONE")
        sub_list.append(last_sub)
    return sub_list

def get_job_tag_check_string(job_tags:list[str], true_script:str="", error_script:str=""):
    script = "if "
    for index, job_tag in enumerate(job_tags):
        script += "[ -f {} ]".format(job_tag)
        if index < len(job_tags)-1:
            script += " && "
    script += "; then\n"
    script += true_script
    script += "else\n"
    script += error_script
    script += "fi\n"
    return script


'''
description: 
make slurm job content
   run_cmd_template = "mpirun -np {} PWmat".format(cpu_per_node or gpu_per_node)
return {*}
author: wuxingxing
'''
def set_slurm_script_content(
                            number_node, 
                            gpu_per_node, #None
                            cpu_per_node,
                            queue_name,
                            custom_flags,
                            env_script,
                            job_name,
                            run_cmd_template,
                            group:list[str],
                            job_tag:str,
                            task_tag:str,
                            task_tag_faild:str,
                            parallel_num:int=1,
                            check_type:str=None
                            ):
        # set head
        script = ""
        if gpu_per_node is None or gpu_per_node == 0:
            script += CPU_SCRIPT_HEAD.format(job_name, number_node, cpu_per_node, queue_name)
            # script += "export CUDA_VISIBLE_DEVICES=''\n"
        else:
            script += GPU_SCRIPT_HEAD.format(job_name, number_node, cpu_per_node, gpu_per_node, queue_name)
        
        for custom_flag in custom_flags:
            script += custom_flag + "\n"
        
        # set conda env
        script += "\n"
        # script += "\n"

        script += "echo \"SLURM_SUBMIT_DIR is $SLURM_SUBMIT_DIR\"\n\n"
        script += "echo \"Starting job $SLURM_JOB_ID at \" `date`\n\n"
        script += "echo \"Running on nodes: $SLURM_NODELIST\"\n\n"

        script += "start=$(date +%s)\n"

        # set source_list
        script += env_script
        script += "\n"
        
        job_cmd = ""
        job_id = 0
        job_tag_list = []
        
        if check_type == DFT_STYLE.pwmat:
            check_info = pwmat_check_success(task_tag, task_tag_faild)
        else:
            check_info = common_check_success(task_tag, task_tag_faild)
            
        while job_id < len(group):
            if group[job_id] == "NONE":
                job_id += 1
                continue
            
            for i in range(parallel_num):
                if group[job_id] == "NONE":
                    job_id += 1
                    continue
                job_cmd += "{\n"
                job_cmd += "cd {}\n".format(group[job_id])
                job_cmd += "if [ ! -f {} ] ; then\n".format(task_tag)
                job_cmd += "    {}\n".format(run_cmd_template)
                job_cmd += check_info
                job_cmd += "fi\n"
                job_cmd += "} &\n\n"
                job_tag_list.append(os.path.join(group[job_id], task_tag))
                job_id += 1
                if job_id > len(group):
                    break
            job_cmd += "wait\n\n"
        
        script += job_cmd
        script += "echo \"Job $SLURM_JOB_ID done at \" `date`\n\n"

        right_script = ""
        right_script += "    end=$(date +%s)\n"
        right_script += "    take=$(( end - start ))\n"
        right_script += "    echo Time taken to execute commands is ${{take}} seconds > {}\n".format(job_tag)
        right_script += "    exit 0\n"   
        error_script  = "    exit 1\n"
        
        script += get_job_tag_check_string(job_tag_list, right_script, error_script)


        return script
    

def pwmat_check_success(task_tag:str, task_tag_faild:str):
    script  = ""
    script += "    if [ -f REPORT ]; then\n"
    script += "        last_line=$(tail -n 1 REPORT)\n"
    script += "        if [[ $last_line == *\"time\"* ]]; then\n"
    script += "            touch {}\n".format(task_tag)
    script += "        else\n"
    script += "            touch {}\n".format(task_tag_faild)
    script += "        fi\n"
    script += "    else\n"
    script += "        touch {}\n".format(task_tag_faild)
    script += "    fi\n"
    return script

def common_check_success(task_tag:str, task_tag_failed:str):
    script = ""
    script += "    if test $? == 0; then\n"
    script += "        touch {}\n".format(task_tag)
    script += "    else\n"
    script += "        touch {}\n".format(task_tag_failed)
    script += "    fi\n"
    return script



from pwact.utils.json_operation import get_parameter, get_required_parameter
from pwact.utils.constant import AL_WORK, DFT_STYLE, SLURM_OUT, CP2K, LAMMPS

class Resource(object):
    # _instance = None
    # scf_style for init_bulk relabel
    def __init__(self, json_dict:dict, job_type:str=AL_WORK.run_iter, dft_style:str=None, scf_style:str=None) -> None:
        if job_type == AL_WORK.run_iter:
            self.train_resource = self.get_resource(get_required_parameter("train", json_dict), default_groupsize=1)
            if self.train_resource.number_node > 1:
                self.train_resource.number_node = 1
            if self.train_resource.gpu_per_node > 1:
                self.train_resource.gpu_per_node = 1
            if self.train_resource.cpu_per_node > 1:
                self.train_resource.cpu_per_node = 1
            print("Warining: the resouce of node, gpu per node and cpu per node  in training automatically adjust to [1, 1, 1]")
            self.train_resource.command = self.train_resource.command.upper()
             
            self.explore_resource = self.get_resource(get_required_parameter("explore", json_dict), default_groupsize=1)
            if "-in" in self.explore_resource.command:
                self.explore_resource.command = self.explore_resource.command.split('-in')[0].strip()
            self.explore_resource.command = "{} -in {} > {}".format(self.explore_resource.command, LAMMPS.input_lammps, SLURM_OUT.md_out)
        else:
            if "explore" in json_dict.keys():
                self.explore_resource = self.get_resource(get_required_parameter("explore", json_dict), default_groupsize=1)
            else:
                self.explore_resource = None
        # check dft resource
        if "dft" in json_dict.keys():
            self.dft_resource = self.get_resource(get_required_parameter("dft", json_dict), default_groupsize=-1)
        else:
            self.dft_resource = ResourceDetail("mpirun -np 1 PWmat", 1, 1, 1, 1, 1, None, None, None)

        if "direct" in json_dict.keys():
            self.direct_resource = self.get_resource(get_required_parameter("direct", json_dict), default_groupsize=1)
        else:
            self.direct_resource = None

        if "scf" in json_dict.keys():
            self.scf_resource = self.get_resource(get_parameter("scf", json_dict, None), default_groupsize=-1) 
        else:
            self.scf_resource = None
        # dftb_command = get_parameter("dftb_command", json_dict["dft"], None)
        # if dftb_command is not None:
        #     self.dft_resource.dftb_command  = "{} > {}".format(dftb_command, SLURM_OUT.dft_out)
        self.dft_style = dft_style
        self.scf_style = scf_style
        if DFT_STYLE.vasp.lower() == dft_style:
            self.dft_resource.command = "{} > {}".format(self.dft_resource.command, SLURM_OUT.dft_out)
        elif DFT_STYLE.pwmat.lower() == dft_style:
            self.dft_resource.command = "{} > {}".format(self.dft_resource.command, SLURM_OUT.dft_out)
        elif DFT_STYLE.cp2k.lower() == dft_style:
            self.dft_resource.command = "{} {} > {}".format(self.dft_resource.command, CP2K.cp2k_inp, SLURM_OUT.dft_out)
        
        if self.scf_resource is not None and scf_style is not None:
            if DFT_STYLE.vasp.lower() == scf_style.lower():
                self.scf_resource.command = "{} > {}".format(self.scf_resource.command, SLURM_OUT.dft_out)
            elif DFT_STYLE.pwmat.lower() == scf_style.lower():
                self.scf_resource.command = "{} > {}".format(self.scf_resource.command, SLURM_OUT.dft_out)
            elif DFT_STYLE.cp2k.lower() == scf_style.lower():
                self.scf_resource.command = "{} {} > {}".format(self.scf_resource.command, CP2K.cp2k_inp, SLURM_OUT.dft_out)

    # @classmethod
    # def get_instance(cls, json_dict:dict = None):
    #     if not cls._instance:
    #         cls._instance = cls(json_dict)
    #     return cls._instance
    
    def get_resource(self, json_dict:dict, default_groupsize=1):
        command = get_required_parameter("command", json_dict)
        group_size = get_parameter("group_size", json_dict, default_groupsize)
        parallel_num = get_parameter("parallel_num", json_dict, 1)
        number_node = get_parameter("number_node", json_dict, 1)
        gpu_per_node = get_parameter("gpu_per_node", json_dict, 0)
        cpu_per_node = get_parameter("cpu_per_node", json_dict, 1)
        queue_name = get_required_parameter("queue_name", json_dict)
        queue_name = queue_name.replace(" ","")
        custom_flags = get_parameter("custom_flags", json_dict, [])
        source_list = get_parameter("source_list", json_dict, [])
        module_list = get_parameter("module_list", json_dict, [])
        env_list = get_parameter("env_list", json_dict, [])
        for i in range(len(custom_flags)):
            if "#SBATCH".lower() not in custom_flags[i].lower():
                custom_flags[i] = "#SBATCH {}".format(custom_flags[i])
            
        env_script = ""
        if len(source_list) > 0:
            for source in source_list:
                if "source" != source.split()[0].lower() and \
                    "export" != source.split()[0].lower() and \
                        "module" != source.split()[0].lower() and \
                            "conda" != source.split()[0].lower():
                    tmp_source = "source {}\n".format(source)
                else:
                    tmp_source = "{}\n".format(source)
                env_script += tmp_source

        if len(module_list) > 0:
            for source in module_list:
                if "module" != source.split()[0].lower():
                    tmp_source = "module load {}\n".format(source)
                else:
                    tmp_source = "{}\n".format(source)
                env_script += tmp_source

        if len(env_list) > 0:
            for source in env_list:
                env_script += source + "\n"

        resource = ResourceDetail(command, group_size, parallel_num, number_node, gpu_per_node, cpu_per_node, queue_name, custom_flags, env_script)
        return resource

class ResourceDetail(object):
    def __init__(self, command:str, group_size:int , parallel_num:int, number_node:int , gpu_per_node:int , cpu_per_node:int ,\
                  queue_name:str, custom_flags:list[str], env_script:str) -> None:
        self.command = command
        self.group_size = group_size
        self.parallel_num = parallel_num
        self.number_node = number_node
        self.gpu_per_node = gpu_per_node
        self.cpu_per_node = cpu_per_node
        self.queue_name = queue_name
        self.custom_flags = custom_flags
        self.env_script = env_script

        if self.gpu_per_node is None and self.cpu_per_node is None:
            raise Exception("ERROR! Both CPU and GPU resources are not specified!")
        # check param
        if "$SLURM".lower() in command.lower():
            pass
        else:
            if "mpirun -np" in command:
                np_num = command.split()[2]
                try:
                    np_num = int(np_num)
                    if np_num > cpu_per_node:
                        raise Exception("the 'command' in resource.json {} set error! The nums of np can not be bigger than 'cpu_per_node'!".format(command))
                except Exception:
                    raise Exception("the 'command' in resource.json {} set error! The nums of np can not be parsed!".format(command))


from maml.sampling.direct import DIRECTSampler, BirchClustering, SelectKFromClusters
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
from ase.io import read
import subprocess, os, sys

write_file = "select.xyz"
if os.path.exists(write_file):
    os.remove(write_file)
filenames = ["candidate.xyz"]
k = 1
threshold = .04
def load_ase_MD_traj(filenames: list):
    """
    Load .traj to pymatgen structures
    """
    structs = []
    trajs = []
    lens = []
    for filename in filenames:
        traj = read(filename,index=":")
        structs += [i for i in traj]
        trajs.append(traj)
        lens.append(len(traj))
    return structs, trajs, lens

structures, trajs, lens = load_ase_MD_traj(filenames)
n_image = len(structures)

DIRECT_sampler = DIRECTSampler(
    clustering=BirchClustering(n=None, threshold_init=threshold), select_k_from_clusters=SelectKFromClusters(k=k)
)

DIRECT_selection = DIRECT_sampler.fit_transform(structures)
n, m = DIRECT_selection["PCAfeatures"].shape

explained_variance = DIRECT_sampler.pca.pca.explained_variance_ratio_
DIRECT_selection["PCAfeatures_unweighted"] = DIRECT_selection["PCAfeatures"] / explained_variance[:m]

plt.plot(
    range(1, explained_variance.shape[0]+1),
    explained_variance * 100,
    "o-",
)
plt.xlabel("i$^{\mathrm{th}}$ PC", size=20)
plt.ylabel("Explained variance", size=20)
ax = plt.gca()
ax.yaxis.set_major_formatter(mtick.PercentFormatter())
plt.tight_layout()
plt.savefig("PCA_variance.png",dpi=360)
plt.close()

def plot_PCAfeature_coverage(all_features, selected_indexes, method="DIRECT"):
    fig, ax = plt.subplots(figsize=(5, 5))
    selected_features = all_features[selected_indexes]
    plt.plot(all_features[:, 0], all_features[:, 1], "*", alpha=0.5, label=f"All {len(all_features):,} structures")
    plt.plot(
        selected_features[:, 0],
        selected_features[:, 1],
        "*",
        alpha=0.5,
        label=f"{method} sampled {len(selected_features):,}",
    )
    legend = plt.legend(frameon=False, fontsize=14, loc="upper left", bbox_to_anchor=(-0.02, 1.02), reverse=True)
    #for lh in legend.legendHandles:
    #    lh.set_alpha(1)
    plt.ylabel("PC 2", size=20)
    plt.xlabel("PC 1", size=20)

all_features = DIRECT_selection["PCAfeatures_unweighted"]
selected_indexes = DIRECT_selection["selected_indexes"]
plot_PCAfeature_coverage(all_features, selected_indexes)
plt.tight_layout()
plt.savefig("PCA_direct.png",dpi=360)
plt.close()

#manual_selection_index = np.arange(0, n_image, int(n_image/n))
#plot_PCAfeature_coverage(all_features, manual_selection_index, "Manually")
#plt.tight_layout()
#plt.savefig("PCA_manually.png",dpi=360)
#plt.close()

def calculate_feature_coverage_score(all_features, selected_indexes, n_bins=100):
    selected_features = all_features[selected_indexes]
    n_all = np.count_nonzero(
        np.histogram(all_features, bins=np.linspace(min(all_features), max(all_features), n_bins))[0]
    )
    n_select = np.count_nonzero(
        np.histogram(selected_features, bins=np.linspace(min(all_features), max(all_features), n_bins))[0]
    )
    return n_select / n_all

def calculate_all_FCS(all_features, selected_indexes, b_bins=100):
    select_scores = [
        calculate_feature_coverage_score(all_features[:, i], selected_indexes, n_bins=b_bins)
        for i in range(all_features.shape[1])
    ]
    return select_scores

all_features = DIRECT_selection["PCAfeatures_unweighted"]
scores_DIRECT = calculate_all_FCS(all_features, DIRECT_selection["selected_indexes"], b_bins=100)
#scores_MS = calculate_all_FCS(all_features, manual_selection_index, b_bins=100)
x = np.arange(len(scores_DIRECT))
x_ticks = [f"PC {n+1}" for n in range(len(x))]

plt.figure(figsize=(15, 4))
plt.bar(
    x,
    scores_DIRECT,
    width=0.3,
    label=f"DIRECT, $\overline{{\mathrm{{Coverage\ score}}}}$ = {np.mean(scores_DIRECT):.3f}",
)
#plt.bar(
#    x + 0.3, scores_MS, width=0.3, label=f"Manual, $\overline{{\mathrm{{Coverage\ score}}}}$ = {np.mean(scores_MS):.3f}"
#)
plt.xticks(x, x_ticks, size=16)
plt.yticks(np.linspace(0, 1.0, 6), size=16)
plt.ylabel("Coverage score", size=20)
plt.legend(shadow=True, loc="lower right", fontsize=16)
plt.tight_layout()
plt.savefig("Cov_score.png",dpi=360)
plt.close()

def get2index(num: int, list_lens: list):
    for idx, i in enumerate(list_lens):
        if num >= i:
            num -= i
        else:
            break
    return idx, num

indices = DIRECT_selection["selected_indexes"]
select_idx = []
for ii,index in enumerate(indices):
    idx, num = get2index(index, lens)
    atoms = trajs[idx][num]
    angles = atoms.cell.cellpar()[-3:]
    if angles.max() > 140 or angles.min() < 40:
        continue
    else:
        atoms.set_scaled_positions(atoms.get_scaled_positions())
        atoms.write(write_file,format="extxyz",append=True)
        select_idx.append(idx)
np.savetxt("select_idx.dat",np.array(indices),fmt="%8d")



import os
import subprocess
import glob
import shutil
import re
import json
import numpy as np
import random
import torch

def check_model_type(model_load_path):
    try:
        _model_checkpoint = torch.load(model_load_path, map_location=torch.device("cpu"))
        model_type = _model_checkpoint['json_file']['model_type']
    except Exception as e:
        with open(model_load_path, 'r') as rf:
            line = rf.readline()
        if "nep" not in line:
            raise Exception("ERROR! The input model file cannot be parsed!")
        else:
            model_type  = "NEP"
    return model_type

'''
description: 
save json_dict to save_path, if file dir is not exist, create it.
param {dict} json_dict
param {*} save_path
return {*}
author: wuxingxing
'''
def save_json_file(json_dict:dict, save_path):
    if not os.path.exists(os.path.dirname(save_path)):
        os.makedirs(os.path.dirname(save_path))
    json.dump(json_dict, open(save_path, "w"), indent=4)

"""
@Description :
读取最后一行文件
@Returns     :
@Author       :wuxingxing
"""
def file_read_last_line(file_path, type_name="int"):
    if os.path.exists(file_path):
        with open(file_path, 'r') as rf:
            last_line = rf.readlines()[-1]  #the last line
            if '[]' in last_line:
                return []
            last_line = last_line.replace(" ","").split(',')
    if len(last_line) > 0 and type_name == "int":
        last_line = [int(i) for i in last_line]
    if len(last_line) > 0 and type_name == "float":
        last_line = [float(i) for i in last_line]
    return last_line

'''
description: 
param {*} file_path
param {*} data_type: 'float' or 'int'
return {*}
author: wuxingxing
'''
def file_read_lines(file_path, data_type="float"):
    data = []
    if os.path.exists(file_path):
        with open(file_path, 'r') as rf:
            lines = rf.readlines()  #the last line
            for line in lines:
                line = re.sub('[\[\]\\n]','',line)
                if len(line) > 0 and data_type == "int":
                    line = [int(i) for i in line.split(',')]
                if len(line) > 0 and data_type == "float":
                    line = [float(i) for i in line.split(',')]
                if len(line) == 1:
                    data.append(line[0])
                else:
                    data.append(line)
    return data

'''
description: 
    load txt data: the content as :
#    step       avg_devi_f       min_devi_f       max_devi_f       avg_devi_e       min_devi_e       max_devi_e
        0      0.016176053      0.000004773      0.023921006      0.494110683      0.007786004      0.680908664
       10      0.020843630      0.000576891      0.033193308      0.481147298      0.001424875      0.696843965
       20      0.020738273      0.001127881      0.031809715      0.478043120      0.001433802      0.703810561
        ...
param {str} file_path
param {*} skiprows
return {*}
author: wuxingxing
'''
def read_data(file_path:str, skiprows=0):
    datas = np.loadtxt(file_path,skiprows=skiprows)
    return datas
    
'''
description: 
 save line str to file_path
param {*} file_path
param {*} line
param {*} mode :'a' or 'w', default is 'a'
return {*}
author: wuxingxing
'''
def write_to_file(file_path, line, mode='w'):
    if not os.path.exists(os.path.dirname(file_path)):
        os.makedirs(os.path.dirname(file_path))
    with open(file_path, mode) as wf:
        wf.write(line)

'''
description: 
    merge file list to one file
param {list} file_list
param {str} save_file
return {*}
author: wuxingxing
'''
def merge_files_to_one(file_list:list[str], save_file:str):
    with open(save_file, 'w') as wf:
        for index, file in enumerate(file_list):
            with open(file, 'r') as rf:
                content = rf.read()
            wf.write(content)
            # if index < len(file_list) - 1:
            #     wf.write("\n")
        
'''
description: 
    copy file from source to target
    If follow_symlinks is not set and src is a symbolic link, a new
    symlink will be created instead of copying the file it points to.
param {str} source_file
param {str} target_file
param {bool} follow_symlinks
return {*}
author: wuxingxing
'''
def copy_file(source_file:str, target_file:str, follow_symlinks:bool=True):
    if not os.path.exists(os.path.dirname(target_file)):
        os.makedirs(os.path.dirname(target_file))
    shutil.copyfile(source_file, target_file, follow_symlinks=follow_symlinks)

'''
description: 
    copy dir to target dir, if target dir is exists, delete it
    symlinks is True: for link type files, maintain link format, else copy as file.
param {str} source_dir
param {str} target_dir
param {*} symlinks
return {*}
author: wuxingxing
'''
def copy_dir(source_dir:str, target_dir:str, symlinks=True):
    if os.path.exists(target_dir):
        if os.path.islink(target_dir) or os.path.isfile(target_dir):
            os.remove(target_dir)
        else:
            shutil.rmtree(target_dir)
    if not os.path.exists(os.path.dirname(target_dir)):
        os.makedirs(os.path.dirname(target_dir))
    shutil.copytree(source_dir, target_dir, symlinks=symlinks)
    
'''
description: 
 link source file to target file, if target file is exist, replace it.
param {str} source_file
param {str} target_file
return {*}
author: wuxingxing
'''
def link_file(source_file:str, target_file:str):
    if os.path.islink(target_file):
        os.remove(target_file)
    if os.path.isfile(target_file):
        os.remove(target_file)
    os.symlink(source_file, target_file)
        
"""
@Description :
删除指定目录下所有文件(该目录不删除) / 或者删除指定文件名文件
@Returns     :
@Author       :wuxingxing
"""

def del_file(path_dir):
    if os.path.exists(path_dir) is False:
        return

    if os.path.isfile(path_dir) or os.path.islink(path_dir):
        os.remove(path_dir)
        return

    for i in os.listdir(path_dir) :
        file_path = os.path.join(path_dir, i)
        if os.path.isfile(file_path) is True:#os.path.isfile判断是否为文件,如果是文件,就删除.如果是文件夹.递归给del_file.
            os.remove(file_path)
        else:
            del_file(file_path)

def del_file_list(path_list:str):
    for _path in path_list:
        del_file(_path)
        if os.path.isdir(_path):
            shutil.rmtree(_path)
        
def del_file_list_by_patten(del_dir:str, patten:str):
    file_list = glob.glob(os.path.join(del_dir, "*{}".format(patten)))
    del_file_list(file_list)

'''
description: 
    mv files under target_dir to source_dir
param {str} source_dir
param {str} target_dir
return {*}
author: wuxingxing
'''
def mv_file(source_file:str, target_file:str):
    if not os.path.exists(source_file):
        return
    if os.path.islink(target_file):
        os.remove(target_file)
    if os.path.exists(target_file):
        shutil.rmtree(target_file)
    shutil.move(source_file, target_file)

def file_shell_op(shell_cmd: str):
    subprocess.call(shell_cmd, shell=True)

'''
description: 
    add postfix to dir, example:
        for .../path/train:
            after add is:
            ../path/train-1-bk, ../path/train-2-bk, ...
param {str} source_dir
param {str} postfix_str
return {*}
author: wuxingxing
'''
def add_postfix_dir(source_dir:str, postfix_str:str):
    index = 1
    while True:
        target_dir = os.path.join(os.path.dirname(source_dir), "{}-{}-{}".format(os.path.basename(source_dir), postfix_str, index))
        if os.path.exists(target_dir):
            index += 1
        else:
            break
    return target_dir


# '''
# description: 
# param {str} source_file
# param {str} target_file
# return {*}
# author: wuxingxing
# '''
# def mv_file(source_file:str, target_file:str):
#     if os.path.islink(target_file):
#         os.remove(target_file)
#     if os.path.exists(target_file):
#         shutil.rmtree(target_file)
#     shutil.move(source_file, target_file)
    
'''
description: 
    delete dir or file if exists.
param {str} del_dir
return {*}
author: wuxingxing
'''
def del_dir(del_dir:str):
    if os.path.exists(del_dir):
        shutil.rmtree(del_dir)

'''
description: 
    Use wildcard characters to search for files in directory search_root_dir that match the wildcard characters
param {str} root_dir
param {str} current_itername
return {*}
author: wuxingxing
'''
def search_files(search_root_dir:str, template:str):
    file_list = glob.glob(os.path.join(search_root_dir, template))
    return file_list

def search_file_by_shell(search_patten:str):
    result = subprocess.run(["ls"], capture_output=True, text=True)
    filtered = [line for line in result.stdout.split("\n") if "pattern" in line]
    return filtered

def str_list_format(input_value):
    input_list = []
    if isinstance(input_value, str):
        input_value = input_value.replace(",", " ")
        input_value = input_value.replace(";", " ")
        input_list = input_value.strip().split()
    elif isinstance(input_value, list):
        input_list = input_value
    else:
        raise Exception("Error format of {}, it should be string or list!".format(input))
    return input_list
        
'''
description: 
    get file extension
param {str} file_name
param {*} split_char default is '.'
return {*}
author: wuxingxing
'''
def get_file_extension(file_name:str, split_char = "."):
    return file_name.split(split_char)[-1].strip()

"""
@Description :
    randomly generate n different nums of int type in the range of [start, end)
@Returns     :
@Author       :wuxingxing
"""

def get_random_nums(start, end, n, seed=None):
    if seed is not None:
        local_random = random.Random(seed)  # 独立的随机实例
        numsArray = set()
        while len(numsArray) < n:
            numsArray.add(local_random.randint(start, end-1))
        return list(numsArray)
    else:
        numsArray = set()
        while len(numsArray) < n:
            numsArray.add(random.randint(start, end-1))
        return list(numsArray)        



"""
    do md with bigmodel
    --/work_path/
    ------------/relax
    -----------------/*_init_config/*_init_.config, etot.input, pseudo files, final.config, mvms_relax
    ------------/super_cell_scale
    -----------------/*_init_config/scale.config, from final.config or *_init_.config
    ------------/perturb
    --------------------/*_init_config/scale.config or final.config
    ----------------------------------/structures/*.config_perturb
    
    ------------/AIMD
    -----------------/*_init_config/*_config_dir the config file from pertub or *_init_.config
    -------------------------------/etot.input, atom.config, ...
    or
    ------------/BIGMODEL
    -----------------/*_init_config/*_config_dir the config file from pertub or *_init_.config
    -------------------------------/bigmodel_md_script.py, POSCAR, ...
    ------------/DIRECT
    -----------------/*_init_config/*_config_dir the config file from pertub or *_init_.config
    -------------------------------/direct_script.py, POSCAR, ...

    -----------/result
    -----------------/*_init_config/relax_mvm_*_init_config if need; aimd_mvm_*_init_config if need
    
"""
import os

from pwact.active_learning.user_input.resource import Resource
from pwact.active_learning.user_input.init_bulk_input import InitBulkParam
from pwact.active_learning.init_bulk.duplicate_scale import get_config_files_with_order

from pwact.utils.constant import PWMAT, INIT_BULK, TEMP_STRUCTURE, SLURM_OUT, DFT_STYLE
from pwact.active_learning.slurm.slurm import SlurmJob, Mission
from pwact.utils.slurm_script import get_slurm_job_run_info, split_job_for_group, set_slurm_script_content
    
from pwact.utils.file_operation import write_to_file, link_file, del_file_list_by_patten, copy_file
from pwact.utils.app_lib.common import link_pseudo_by_atom, set_input_script
from pwact.data_format.configop import save_config, get_atom_type

class DIRECT(object):
    def __init__(self, resource: Resource, input_param:InitBulkParam):
        self.resource = resource
        self.input_param = input_param
        self.init_configs = self.input_param.sys_config
        self.relax_dir = os.path.join(self.input_param.root_dir, TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.relax)
        self.super_cell_scale_dir = os.path.join(self.input_param.root_dir, TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.super_cell_scale)
        self.pertub_dir = os.path.join(self.input_param.root_dir,TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.pertub)
        self.bigmodel_dir = os.path.join(self.input_param.root_dir, TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.bigmodel)
        self.real_bigmodel_dir = os.path.join(self.input_param.root_dir, INIT_BULK.bigmodel)
        self.direct_dir = os.path.join(self.input_param.root_dir, TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.direct)
        self.real_direct_dir = os.path.join(self.input_param.root_dir, INIT_BULK.direct)
  
    def make_bigmodel_work(self):
        bigmodel_paths = []
        for init_config in self.init_configs:
            if init_config.bigmodel is False:
                continue
            init_config_name = "init_config_{}".format(init_config.config_index)
            # config_list, config_type = get_config_files_with_order(self.super_cell_scale_dir, self.relax_dir, init_config_name, init_config.config, self.pertub_dir)
            config_list, config_type = get_config_files_with_order(
            super_cell_scale_dir=self.super_cell_scale_dir,
            relax_dir=self.relax_dir,
            init_config_dirname=init_config_name, 
            init_config_path=init_config.config_file, 
            pertub_dir=self.pertub_dir,
            dft_style=self.resource.dft_style
            )

            for index, config in enumerate(config_list):
                if config_type == INIT_BULK.pertub:
                    tmp_config_dir = os.path.basename(os.path.basename(os.path.dirname(config)))
                elif config_type == INIT_BULK.super_cell or config_type == INIT_BULK.scale:
                    tmp_config_dir = os.path.basename(config).replace(DFT_STYLE.get_postfix(self.resource.dft_style), "")
                elif config_type == INIT_BULK.relax:
                    tmp_config_dir = INIT_BULK.relax
                else:
                    tmp_config_dir = INIT_BULK.init
                bigmodel_dir = os.path.join(self.bigmodel_dir, init_config_name, tmp_config_dir, "{}_{}".format(index, INIT_BULK.bigmodel))
                if not os.path.exists(bigmodel_dir):
                    os.makedirs(bigmodel_dir)
                
                self.make_bigmodel_file(
                    bigmodel_dir=bigmodel_dir, 
                    config_file=config, 
                    conifg_format=DFT_STYLE.get_format_by_postfix(os.path.basename(config)), 
                    input_file=init_config.bigmodel_input_file)

                bigmodel_paths.append(bigmodel_dir)
        # make slurm script and slurm job
        self.make_bigmodel_slurm_job_files(bigmodel_paths)
    
    def check_work_done(self):
        slurm_remain, slurm_success = get_slurm_job_run_info(self.bigmodel_dir, \
            job_patten="*-{}".format(INIT_BULK.bigmodel_job), \
            tag_patten="*-{}".format(INIT_BULK.bigmodel_tag))
        slurm_done = True if len(slurm_remain) == 0 and len(slurm_success) > 0 else False # len(slurm_remain) > 0 exist slurm jobs need to do
        return slurm_done
            
    def do_bigmodel_jobs(self):
        mission = Mission()
        slurm_remain, slurm_success = get_slurm_job_run_info(self.bigmodel_dir, \
            job_patten="*-{}".format(INIT_BULK.bigmodel_job), \
            tag_patten="*-{}".format(INIT_BULK.bigmodel_tag))
        slurm_done = True if len(slurm_remain) == 0 and len(slurm_success) > 0 else False # len(slurm_remain) > 0 exist slurm jobs need to do
        if slurm_done is False:
            #recover slurm jobs
            if len(slurm_remain) > 0:
                print("Run these bigmodel Jobs:\n")
                print(slurm_remain)
                for i, script_path in enumerate(slurm_remain):
                    slurm_job = SlurmJob()
                    tag_name = "{}-{}".format(os.path.basename(script_path).split('-')[0].strip(), INIT_BULK.bigmodel_tag)
                    tag = os.path.join(os.path.dirname(script_path),tag_name)
                    slurm_job.set_tag(tag)
                    slurm_job.set_cmd(script_path)
                    mission.add_job(slurm_job)

            if len(mission.job_list) > 0:
                mission.commit_jobs()
                mission.check_running_job()
                mission.all_job_finished()
                # mission.move_slurm_log_to_slurm_work_dir()

    '''
    description: 
        input_file is aimd input control file, for vasp is incar, for pwmat is etot.input
    return {*}
    author: wuxingxing
    '''    
    def make_bigmodel_file(
            bigmodel_dir=str, 
            config_file=str, 
            conifg_format=str, 
            input_file=str):
        #1. set config file, cvt to poscar format
        target_config = save_config(config=config_file, 
                            input_format=conifg_format,
                            wrap = False, 
                            direct = True, 
                            sort = True, 
                            save_format=DFT_STYLE.poscar, 
                            save_path=bigmodel_dir, 
                            save_name=DFT_STYLE.get_normal_config(DFT_STYLE.vasp)
                            )

        atom_type_list, _ = get_atom_type(config_file, conifg_format)
        #2. copy script file
        copy_file(input_file, os.path.join(bigmodel_dir, os.path.basename(input_file)))

    def make_bigmodel_slurm_job_files(self, bigmodel_dir_list:list[str]):
        del_file_list_by_patten(self.bigmodel_dir, "*{}".format(INIT_BULK.bigmodel_job))
        group_list = split_job_for_group(self.resource.explore_resource.group_size, bigmodel_dir_list, self.resource.explore_resource.parallel_num)
        for group_index, group in enumerate(group_list):
            if group[0] == "NONE":
                continue
            jobname = "bigmodel{}".format(group_index)
            tag_name = "{}-{}".format(group_index, INIT_BULK.bigmodel_tag)
            tag = os.path.join(self.bigmodel_dir, tag_name)
            run_cmd = self.resource.explore_resource.command
            group_slurm_script = set_slurm_script_content(gpu_per_node=self.resource.explore_resource.gpu_per_node, 
                number_node = self.resource.explore_resource.number_node, 
                cpu_per_node = self.resource.explore_resource.cpu_per_node,
                queue_name = self.resource.explore_resource.queue_name,
                custom_flags = self.resource.explore_resource.custom_flags,
                env_script = self.resource.explore_resource.env_script,
                job_name = jobname,
                run_cmd_template = run_cmd,
                group = group,
                job_tag = tag,
                task_tag = INIT_BULK.bigmodel_tag, 
                task_tag_faild = INIT_BULK.bigmodel_tag_failed,
                parallel_num=self.resource.dft_resource.parallel_num,
                check_type=self.resource.dft_style
                )
            slurm_script_name = "{}-{}".format(group_index, INIT_BULK.bigmodel_job)
            slurm_job_file =  os.path.join(self.bigmodel_dir, slurm_script_name)
            write_to_file(slurm_job_file, group_slurm_script, "w")

    def do_post_process(self):
        if os.path.exists(self.bigmodel_dir):
            link_file(self.bigmodel_dir, self.real_bigmodel_dir)



"""
    realx dor strcuture
    --/work_path/
    ------------/relax
    -----------------/*_init_config/*_init_.config, etot.input, pseudo files, final.config, mvms_relax
    ------------/super_cell_scale
    -----------------/*_init_config/scale.config, from final.config or *_init_.config
    ------------/perturb
    --------------------/*_init_config/scale.config or final.config
    ----------------------------------/structures/*.config_perturb
    ------------/AIMD
    -----------------/*_init_config/*_config_dir the config file from pertub or *_init_.config
    -------------------------------/etot.input, atom.config, ...
    -----------/result
    -----------------/*_init_config/relax_mvm_*_init_config if need; aimd_mvm_*_init_config if need
    
"""
import os
from pwact.active_learning.user_input.resource import Resource
from pwact.active_learning.user_input.init_bulk_input import InitBulkParam, Stage

from pwact.active_learning.slurm.slurm import SlurmJob, Mission
from pwact.utils.slurm_script import get_slurm_job_run_info, split_job_for_group, set_slurm_script_content
    
from pwact.utils.constant import PWMAT, INIT_BULK, TEMP_STRUCTURE, SLURM_OUT, DFT_STYLE, PWDATA

from pwact.utils.file_operation import write_to_file, link_file, search_files, mv_file, del_file, copy_file, del_file_list_by_patten
from pwact.utils.app_lib.common import link_pseudo_by_atom, set_input_script
from pwact.data_format.configop import save_config, get_atom_type

class Relax(object):
    def __init__(self, resource: Resource, input_param:InitBulkParam):
        self.resource = resource
        self.input_param = input_param
        self.init_configs = self.input_param.sys_config
        self.relax_dir = os.path.join(self.input_param.root_dir, TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.relax)
        self.relax_real_dir = os.path.join(self.input_param.root_dir, INIT_BULK.relax)
        
    def check_work_done(self):
        slurm_remain, slurm_success = get_slurm_job_run_info(self.relax_dir, \
        job_patten="*-{}".format(INIT_BULK.relax_job), \
        tag_patten="*-{}".format(INIT_BULK.relax_tag))
        slurm_done = True if len(slurm_remain) == 0 and len(slurm_success) > 0 else False # len(slurm_remain) > 0 exist slurm jobs need to do
        return slurm_done
    
    def make_relax_work(self):
        relax_paths = []
        for init_config in self.init_configs:
            if init_config.relax:
                init_config_name = "{}_{}".format(INIT_BULK.init_config, init_config.config_index)
                relax_path = os.path.join(self.relax_dir, init_config_name)
                if not os.path.exists(relax_path):
                    os.makedirs(relax_path)
                self.make_relax_file(relax_path, init_config=init_config)
                relax_paths.append(relax_path)
        # make slurm script and slurm job
        self.make_relax_slurm_job_files(relax_paths)
    
    def do_relax_jobs(self):
        mission = Mission()
        slurm_remain, slurm_success = get_slurm_job_run_info(self.relax_dir, \
            job_patten="*-{}".format(INIT_BULK.relax_job), \
            tag_patten="*-{}".format(INIT_BULK.relax_tag))
        slurm_done = True if len(slurm_remain) == 0 and len(slurm_success) > 0 else False # len(slurm_remain) > 0 exist slurm jobs need to do
        if slurm_done is False:
            #recover slurm jobs
            if len(slurm_remain) > 0:
                print("Run these relax Jobs:\n")
                print(slurm_remain)
                for i, script_path in enumerate(slurm_remain):
                    slurm_job = SlurmJob()
                    tag_name = "{}-{}".format(os.path.basename(script_path).split('-')[0].strip(), INIT_BULK.relax_tag)
                    tag = os.path.join(os.path.dirname(script_path),tag_name)
                    slurm_job.set_tag(tag)
                    slurm_job.set_cmd(script_path)
                    mission.add_job(slurm_job)

            if len(mission.job_list) > 0:
                mission.commit_jobs()
                mission.check_running_job()
                mission.all_job_finished(error_type=SLURM_OUT.dft_out)
                # mission.move_slurm_log_to_slurm_work_dir()
                
    def make_relax_file(self, relax_path, init_config:Stage):
        #1. set config file
        target_config = save_config(config=init_config.config_file,
                                    input_format=init_config.format,
                                    wrap = False, 
                                    direct = True, 
                                    sort = True, 
                                    save_format=DFT_STYLE.get_pwdata_format(dft_style=self.input_param.dft_style, is_cp2k_coord=True), 
                                    save_path=relax_path, 
                                    save_name=DFT_STYLE.get_normal_config(self.input_param.dft_style))

        atom_type_list, _ = get_atom_type(init_config.config_file, init_config.format)
        #2. set pseudo files
        if not init_config.use_dftb:
            pseudo_names = link_pseudo_by_atom(
                pseudo_list     = self.input_param.dft_input.pseudo, 
                target_dir      = relax_path, 
                atom_order      = atom_type_list, 
                dft_style       = self.input_param.dft_style,
                basis_set_file  =self.input_param.dft_input.basis_set_file,
                potential_file  =self.input_param.dft_input.potential_file)
        else:
            # link in.skf path to aimd dir
            target_dir = os.path.join(relax_path, PWMAT.in_skf)
            if self.input_param.dft_input.in_skf is not None:
                link_file(self.input_param.dft_input.in_skf, target_dir)
            pseudo_names = []
        #3. make input file, for vasp is INCAR, for PWMAT is etot.input, for cp2k is inp file
        set_input_script(
            input_file=init_config.relax_input_file,
            config=target_config,
            kspacing=init_config.relax_kspacing, 
            flag_symm=init_config.relax_flag_symm, 
            dft_style=self.input_param.dft_style,
            save_dir=relax_path,
            pseudo_names=pseudo_names,
            gaussian_base_param=self.input_param.dft_input.gaussian_base_param,# these for cp2k
            # xc_functional=self.input_param.dft_input.xc_functional,
            # potential=self.input_param.dft_input.potential,
            # basis_set=self.input_param.dft_input.basis_set
        )

    def make_relax_slurm_job_files(self, relax_sub_list:list[str]):
        del_file_list_by_patten(self.relax_dir, "*{}".format(INIT_BULK.relax_job))
        group_list = split_job_for_group(self.resource.dft_resource.group_size, relax_sub_list, self.resource.dft_resource.parallel_num)
        for group_index, group in enumerate(group_list):
            if group[0] == "NONE":
                continue
            jobname = "relax{}".format(group_index)
            tag_name = "{}-{}".format(group_index, INIT_BULK.relax_tag)
            tag = os.path.join(self.relax_dir, tag_name)
            # if self.resource.dft_resource.gpu_per_node > 0:
            #     run_cmd = "mpirun -np {} PWmat".format(self.resource.dft_resource.gpu_per_node)
            # else:
            #     raise Exception("ERROR! the cpu version of pwmat not support yet!")
            run_cmd = self.resource.dft_resource.command
            group_slurm_script = set_slurm_script_content(
                number_node = self.resource.dft_resource.number_node, 
                gpu_per_node=self.resource.dft_resource.gpu_per_node, 
                cpu_per_node = self.resource.dft_resource.cpu_per_node,
                queue_name = self.resource.dft_resource.queue_name,
                custom_flags = self.resource.dft_resource.custom_flags,
                env_script = self.resource.dft_resource.env_script,
                job_name = jobname,
                run_cmd_template = run_cmd,
                group = group,
                job_tag = tag,
                task_tag = INIT_BULK.relax_tag, 
                task_tag_faild = INIT_BULK.relax_tag_failed,
                parallel_num=self.resource.dft_resource.parallel_num,
                check_type=self.resource.dft_style
                )
            slurm_script_name = "{}-{}".format(group_index, INIT_BULK.relax_job)
            slurm_job_file = os.path.join(self.relax_dir, slurm_script_name)
            write_to_file(slurm_job_file, group_slurm_script, "w")

    '''
    description: 
     after relax calculate:
        for PWmat: change the 'final.config' to relaxed.config
        for VASP: change the 'CONTCAR' to  relaxed.poscar
        for CP2k: change the 'dft.log' to relaxed.poscar
        
    param {*} self
    return {*}
    author: wuxingxing
    '''    
    def do_post_process(self):
        # no use thie code
        # 1. change the name of relaxed config to relaxed.config for pwmat or relaxed.poscar to vasp
        final_configs = search_files(self.relax_dir, "*/{}".format(DFT_STYLE.get_relaxed_original_name(self.resource.dft_style)))
        for final_config in final_configs:
            target_config = os.path.join(os.path.dirname(final_config), DFT_STYLE.get_relaxed_config(self.resource.dft_style))
            if self.resource.dft_style == DFT_STYLE.cp2k:
                save_config(config=final_config, #convert cp2k scf to poscar format
                        input_format=PWDATA.cp2k_scf, 
                        direct=False, 
                        sort=False, 
                        save_path=os.path.dirname(target_config), 
                        save_name=os.path.basename(target_config),
                        save_format=PWDATA.vasp_poscar)
            else:# rename final.config to relaxed.config, or contcar to relaxed.poscar
                copy_file(final_config, target_config)
                
        # 2. link relax_dir to relax_real_dir
        if not os.path.exists(self.relax_dir):
            link_file(self.relax_dir, self.relax_real_dir)

    def delete_nouse_files(self):
        #1. mv init_config_* to real_dir
        mv_file(self.relax_dir, self.relax_real_dir)
        #2. delete tag and slurm files
        tag_list  = search_files(self.relax_real_dir,  template="*-tag-*")
        slurm_log = search_files(self.relax_real_dir,  template="*slurm-*.out")
        del_file(tag_list)
        del_file(slurm_log)
        
    


from pwdata import Config
import os, sys, glob
import numpy as np

def count_pwdata(work_dir):
    
    dir_list = glob.glob(os.path.join(work_dir, "*"))
    res = []
    for dir in dir_list:
        # train
        train_num = np.load(os.path.join(dir, "train/energies.npy")).shape[0]
        res.append(train_num)

        if os.path.exists(os.path.join(dir, "valid/energies.npy")):
            test_num = np.load(os.path.join(dir, "valid/energies.npy")).shape[0]
            res.append(test_num)
            print("{} {} {}".format( dir, train_num, test_num))
        else:
            print("{} {}".format(dir, train_num))
    print(np.sum(res))

def count_outmlmd():
    work_dir = "/data/home/wuxingxing/datas/debugs/dengjiapei/run_iter/iter.0000/label/scf"
    mlmds = glob.glob(os.path.join(work_dir, "*/*/*/OUT.MLMD"))
    print(len(mlmds))

def save_mlmd():
    work_dir = "/data/home/wuxingxing/datas/debugs/dengjiapei/run_iter"
    data_list = glob.glob(os.path.join(work_dir, "iter.*/label/scf/*/*/*/OUT.MLMD"))
    datasets_path = "/data/home/wuxingxing/datas/debugs/dengjiapei/run_iter/mlmd_pwdata"
    # data_name = datasets_path
    image_data = None
    for data_path in data_list:
        if image_data is not None:
            tmp_config = Config("pwmat/movement", data_path)
            # if not isinstance(tmp_config, list):
            #     tmp_config = [tmp_config]
            image_data.images.extend(tmp_config.images)
        else:
            image_data = Config("pwmat/movement", data_path)
            
            if not isinstance(image_data.images, list):
                image_data.images = [image_data.images]
        
            # if not isinstance(image_data, list):
            #     image_data = [image_data]
    image_data.to(
                output_path=datasets_path,
                save_format="pwmlff/npy",
                train_ratio = 0.8, 
                train_data_path="train", 
                valid_data_path="valid", 
                random=True,
                seed = 2024, 
                retain_raw = False
                )
    print(len(image_data.images))

def find_outcar_files(directory):
    outcar_files = []
    for root, dirs, files in os.walk(directory):
        for file in files:
            if file == 'OUTCAR' or file == "REPORT":
                outcar_files.append(os.path.join(root, file))
            
            
    return outcar_files

def is_convergence(file_path, format):
    def _is_cvg_vasp(file_path:str):
        with open(file_path, 'r') as rf:
            outcar_contents = rf.readlines()
        nelm = None
        ediff = None
        for idx, ii in enumerate(outcar_contents):
            if 'NELM   =' in ii:
                nelm = int(ii.split()[2][:-1])
            if 'EDIFF = ' in ii:
                ediff = float(ii.split()[-1])
        
        with open(os.path.join(os.path.dirname(os.path.abspath(file_path)), "OSZICAR"), 'r') as rf:
            oszi_contents = rf.readlines()
        _split = oszi_contents[-2].split()
        real_nelm = int(_split[1])
        real_ediff1 = abs(float(_split[3]))
        real_ediff2 = abs(float(_split[4]))

        if real_nelm < nelm:
            return True
        elif real_ediff1 <= ediff and real_ediff2 <=ediff:
            return True
        else:
            False

    def _is_cvg_pwmat(file_path:str):
        with open(file_path, 'r') as rf:
            report_contents = rf.readlines()
        e_error   = None
        rho_error = None
        etot_idx = -1
        drho_idx = -1
        for idx, ii in enumerate(report_contents):
            if e_error is None and 'E_ERROR   =' in ii:
                e_error = abs(float(ii.split()[-1]))
            if rho_error is None and 'RHO_ERROR =' in ii:
                rho_error = abs(float(ii.split()[-1]))
            if 'E_tot(eV)            =' in ii:
                etot_idx = idx
            if 'dv_ave, drho_tot     =' in ii:
                drho_idx = idx
            if 'niter reached' in ii:
                break
            elif 'ending_scf_reason = tol' in ii:
                return True

        if e_error >= abs(float(report_contents[etot_idx].split()[-1])) or \
            rho_error >= abs(float(report_contents[drho_idx].split()[-1])):
            return True
        return False
        
    if format == "vasp":
        return _is_cvg_vasp(file_path)
    elif format == "pwmat":
        return _is_cvg_pwmat(file_path)
    elif format == "cp2k":
        return True

def check_convergence(file_path:list[str], format:str):
    cvg_files = []
    uncvg_files = []
    cvg_infos = ""
    cvg_detail_infos=""
    for file in file_path:
        if is_convergence(file, format):
            cvg_files.append(file)
        else:
            uncvg_files.append(file)
    
    cvg_infos += "Number of converged files: {}, number of non-converged files: {}".format(len(cvg_files), len(uncvg_files))
    cvg_detail_infos += cvg_infos
    cvg_detail_infos += "\nList of non-converged files:\n{}".format("\n".join(uncvg_files))
    return cvg_files, uncvg_files, cvg_infos, cvg_detail_infos
    
def cvt_config():
    pwdata = "/data/home/wuxingxing/datas/pwmat_mlff_workdir/auag/pwdata/Ag4Au44"
    image = Config.read(data_path=pwdata, format="pwmlff/npy")[0][0]
    image.to(data_path="/share/public/PWMLFF_test_data/pwact_examples/25-pwact-demo/structures/AuAg", data_name="ag4au44-atom.config", format="pwmat/config")

def remove_package_and_newton_lines(filename):
    import re
    """
    检查文件中是否存在 'package kokkos neigh'（忽略空格和大小写），
    如果存在，则删除该行以及所有 'newton on' 行。
    """
    # 目标字符串去除所有空白并转为小写
    package_pattern = re.compile(r'package\s+kokkos\s+neigh', re.IGNORECASE)
    newton_pattern = re.compile(r'newton\s+on', re.IGNORECASE)

    with open(filename, 'r', encoding='utf-8') as f:
        lines = f.readlines()

    has_package = any(package_pattern.search(line) for line in lines)
    if not (has_package):
        return lines
    
    new_lines = []
    for line in lines:
        if package_pattern.search(line) or newton_pattern.search(line):
            continue
        new_lines.append(line)

    print(new_lines)
    # return new_lines


if __name__=="__main__":
    # count_pwdata(work_dir = "/data/home/wuxingxing/datas/debugs/dengjiapei/run_iter/mlmd_pwdata")
    # count_outmlmd()
    # save_mlmd()
    # cvt_config()

    # os.chdir("/share/public/PWMLFF_test_data/pwact_examples/25-pwact-demo/si_pwmat/run_iter_lmps/iter.0000/temp_run_iter_work/02.label/scf")
    # current_dir = os.getcwd()
    # outcar_files = find_outcar_files(current_dir)
    # cvg_files, uncvg_files, cvg_infos, cvg_detail_infos = check_convergence(outcar_files, "pwmat")
    # print(cvg_detail_infos)
    # print(cvg_infos)

    remove_package_and_newton_lines("/data/home/wuxingxing/datas/al_dir/in_lmps_files/in0.lmps")
    


from pandas.core.frame import DataFrame
import os

import pandas as pd
import numpy as  np
from pwact.utils.constant import EXPLORE_FILE_STRUCTURE, UNCERTAINTY, SLURM_OUT
from pwact.utils.file_operation import write_to_file, search_files, read_data

from pwact.utils.format_input_output import get_sub_md_sys_template_name
def _select_image(save_dir:str, devi_pd:DataFrame, lower:float, higer:float, max_select:float):
    accurate_pd  = devi_pd[devi_pd[EXPLORE_FILE_STRUCTURE.devi_columns[0]] < lower]
    candidate_pd = devi_pd[(devi_pd[EXPLORE_FILE_STRUCTURE.devi_columns[0]] >= lower) & (devi_pd[EXPLORE_FILE_STRUCTURE.devi_columns[0]] < higer)]
    error_pd     = devi_pd[devi_pd[EXPLORE_FILE_STRUCTURE.devi_columns[0]] > higer]
    #4. if selected images more than number limitaions, randomly select
    remove_candi = None
    rand_candi = None
    if candidate_pd.shape[0] > max_select:
        rand_candi = candidate_pd.sample(max_select)
        remove_candi = candidate_pd.drop(rand_candi.index)
    
    #5. save select info
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
    summary = "Total structures {}    accurate {} rate {:.2f}%    selected {} rate {:.2f}%    error {} rate {:.2f}%\n"\
        .format(devi_pd.shape[0], accurate_pd.shape[0], accurate_pd.shape[0]/devi_pd.shape[0]*100, \
                    candidate_pd.shape[0], candidate_pd.shape[0]/devi_pd.shape[0]*100, \
                        error_pd.shape[0], error_pd.shape[0]/devi_pd.shape[0]*100)

    accurate_pd.to_csv(os.path.join(save_dir, EXPLORE_FILE_STRUCTURE.accurate))
    candi_info = ""
    if rand_candi is not None:
        rand_candi.to_csv(os.path.join(save_dir, EXPLORE_FILE_STRUCTURE.candidate))
        remove_candi.to_csv(os.path.join(save_dir, EXPLORE_FILE_STRUCTURE.candidate_delete))
        candi_info += "Candidate configurations: {}, randomly select {}, delete {}\n        Select details in file {}\n        Delete details in file {}.\n".format(
                candidate_pd.shape[0], rand_candi.shape[0], remove_candi.shape[0],\
                EXPLORE_FILE_STRUCTURE.candidate, EXPLORE_FILE_STRUCTURE.candidate_delete)
    else:
        candidate_pd.to_csv(os.path.join(save_dir, EXPLORE_FILE_STRUCTURE.candidate))
        candi_info += "        Candidate configurations: {}\n    Select details in file {}\n".format(
                candidate_pd.shape[0], EXPLORE_FILE_STRUCTURE.candidate)
            
    error_pd.to_csv(os.path.join(save_dir, EXPLORE_FILE_STRUCTURE.failed))
    
    summary_info = ""

    summary_info += summary
    summary_info += "\nSelect by model deviation force:\n"
    summary_info += "Accurate configurations: {}, details in file {}\n".\
        format(accurate_pd.shape[0], EXPLORE_FILE_STRUCTURE.accurate)
        
    summary_info += candi_info
        
    summary_info += "Error configurations: {}, details in file {}\n".\
        format(error_pd.shape[0], EXPLORE_FILE_STRUCTURE.failed)
    
    write_to_file(os.path.join(save_dir, EXPLORE_FILE_STRUCTURE.select_summary), summary_info, "w")
    return summary_info, summary


def sort_model_devi_files(model_devi_files: list):
    md_sys_dict = {}
    for file in model_devi_files:
        md_sys = os.path.basename(os.path.dirname(os.path.dirname(file)))
        sys = int(md_sys.split('.')[3])
        md = int(md_sys.split('.')[1])
        if md not in md_sys_dict.keys():
            md_sys_dict[md] = {sys:[file]}
        else:
            if sys in md_sys_dict[md].keys():
                md_sys_dict[md][sys].append(file)
            else:
                md_sys_dict[md][sys] = [file]
    return md_sys_dict

def select_image(
    md_dir:str, 
    save_dir:str,
    md_job:list,
    devi_name:str,
    lower:float, 
    higer:float
):  
    #1. get model_deviation file
    model_deviation_patten = "{}/{}".format(get_sub_md_sys_template_name(), devi_name)
    model_devi_files = search_files(md_dir, model_deviation_patten)
    model_devi_files = sorted(model_devi_files)
    md_sys_dict = sort_model_devi_files(model_devi_files)
    
    error_pd =None
    accurate_pd =None
    rand_candi =None
    remove_candi =None
    
    error_pd = pd.DataFrame(columns=EXPLORE_FILE_STRUCTURE.devi_columns)
    accurate_pd = pd.DataFrame(columns=EXPLORE_FILE_STRUCTURE.devi_columns)
    rand_candi = pd.DataFrame(columns=EXPLORE_FILE_STRUCTURE.devi_columns)
    remove_candi = pd.DataFrame(columns=EXPLORE_FILE_STRUCTURE.devi_columns)

    for md in md_sys_dict.keys():
        sys_dict = md_sys_dict[md]
        for sys_idx, sys in enumerate(sys_dict.keys()):
            devi_files = sys_dict[sys]
            select_num = md_job[md].select_sys[sys_idx]
            tmp_devi_pd, _base_kpu = read_pd_files(devi_files)
            if len(_base_kpu) > 0: # for kpu upper and lower
                _lower = np.mean(_base_kpu)*lower
                _higer = _lower * higer
            else:
                _lower = lower
                _higer = higer
            tmp_error_pd, tmp_accurate_pd, tmp_rand_candi, tmp_remove_candi = select_pd(tmp_devi_pd, _lower, _higer, select_num)
            error_pd = pd.concat([error_pd, tmp_error_pd])
            accurate_pd = pd.concat([accurate_pd, tmp_accurate_pd])
            rand_candi = pd.concat([rand_candi, tmp_rand_candi])
            remove_candi = pd.concat([remove_candi, tmp_remove_candi])
    right_md, error_md = count_mdstop_info(model_devi_files)
    md_run_info = "A total of {} MD trajectories were run. with {} trajectories correctly executed and {} trajectories normally completed. \nFor detailed information, refer to File {}."\
        .format(len(right_md) + len(error_md), len(right_md), len(error_md), EXPLORE_FILE_STRUCTURE.md_traj_error_record)
    
    summary_info, summary = count_info(save_dir, error_pd, accurate_pd, rand_candi, remove_candi, md_run_info)
    print("Image select result:\n {}\n\n".format(summary_info))

    write_to_file(os.path.join(save_dir, EXPLORE_FILE_STRUCTURE.md_traj_error_record), md_run_info, "w")
    details = "\n"
    if len(error_md) > 0:
        details += "\nUnfinished md trajectory directory:\n"
        details += "\n".join(error_md)
    if len(right_md) > 0:
        details += "\n\nCorrectly run md trajectory directory:\n"
        details += "\n".join(right_md)
    write_to_file(os.path.join(save_dir, EXPLORE_FILE_STRUCTURE.md_traj_error_record), details, "a")

    return summary

def print_select_image(
    md_dir:str, 
    save_dir:str,
    devi_name:str,
    lower:float, 
    higer:float
):
    #1. get model_deviation file
    model_deviation_patten = "{}/{}".format(get_sub_md_sys_template_name(), devi_name)
    model_devi_files = search_files(md_dir, model_deviation_patten)
    model_devi_files = sorted(model_devi_files)
    md_sys_dict = sort_model_devi_files(model_devi_files)
    
    error_pd =None
    accurate_pd =None
    rand_candi =None
    remove_candi =None

    for md in md_sys_dict.keys():
        sys_dict = md_sys_dict[md]
        for sys_idx, sys in enumerate(sys_dict.keys()):
            devi_files = sys_dict[sys]
            tmp_devi_pd, _base_kpu = read_pd_files(devi_files)
            if len(_base_kpu) > 0: # for kpu upper and lower
                _lower = np.mean(_base_kpu)*lower
                _higer = _lower * higer
            else:
                _lower = lower
                _higer = higer
            tmp_error_pd, tmp_accurate_pd, tmp_rand_candi, tmp_remove_candi = select_pd(tmp_devi_pd, _lower, _higer, 10000000)
            error_pd = pd.concat([error_pd, tmp_error_pd]) if error_pd is not None else tmp_error_pd
            accurate_pd = pd.concat([accurate_pd, tmp_accurate_pd]) if error_pd is not None else tmp_accurate_pd
            rand_candi = pd.concat([rand_candi, tmp_rand_candi]) if error_pd is not None else tmp_rand_candi
            remove_candi = pd.concat([remove_candi, tmp_remove_candi]) if error_pd is not None else tmp_remove_candi
    summary_info, summary = count_info(save_dir, error_pd, accurate_pd, rand_candi, remove_candi)
    
    # summary_info, summary = select_image(save_dir=self.select_dir, 
    #                 devi_pd=devi_pd, 
    #                 lower=self.input_param.strategy.lower_model_deiv_f, 
    #                 higer=self.input_param.strategy.upper_model_deiv_f, 
    #                 max_select=self.input_param.strategy.max_select)
    print("Image select result (lower {} upper {}):\n {}\n\n".format(lower, higer, summary_info))
    return summary


def select_pd(devi_pd:DataFrame, lower:float, higer:float, max_select:float):
    accurate_pd  = devi_pd[devi_pd[EXPLORE_FILE_STRUCTURE.devi_columns[0]] < lower]
    candidate_pd = devi_pd[(devi_pd[EXPLORE_FILE_STRUCTURE.devi_columns[0]] >= lower) & (devi_pd[EXPLORE_FILE_STRUCTURE.devi_columns[0]] < higer)]
    error_pd     = devi_pd[devi_pd[EXPLORE_FILE_STRUCTURE.devi_columns[0]] > higer]
    #4. if selected images more than number limitaions, randomly select
    cand_remove_candi = None
    cand_rand_candi = None
    if candidate_pd.shape[0] > max_select:
        cand_rand_candi = candidate_pd.sample(max_select)
        cand_remove_candi = candidate_pd.drop(cand_rand_candi.index)
    else:
        cand_rand_candi = candidate_pd
        cand_remove_candi = pd.DataFrame(columns=cand_rand_candi.columns)
    return error_pd, accurate_pd, cand_rand_candi, cand_remove_candi

def read_pd_files(model_devi_files:list[str]):
    devi_pd = pd.DataFrame()#columns=EXPLORE_FILE_STRUCTURE.devi_columns
    base_force_kpu = []
    if os.path.basename(model_devi_files[0]) == EXPLORE_FILE_STRUCTURE.kpu_model_devi:
        for devi_file in model_devi_files:
                # the data format of devi_file example:
                #     step        etot_kpu      f_kpu_mean       f_kpu_max       f_kpu_min
                #        0            0.00           36.09          107.61            2.00
                #        5            0.00          103.58          327.85            2.00
                # ......
                devi_force = np.loadtxt(devi_file)
                tmp_pd = pd.DataFrame()
                tmp_pd[EXPLORE_FILE_STRUCTURE.devi_columns[0]] = devi_force[:, 2]#
                tmp_len = int(devi_force.shape[0]*0.01)
                tmp_len = 5 if tmp_len < 5 else tmp_len
                base_force_kpu.extend(devi_force[:, 2][:tmp_len])
                tmp_pd[EXPLORE_FILE_STRUCTURE.devi_columns[1]] = devi_force[:, 0]
                tmp_pd[EXPLORE_FILE_STRUCTURE.devi_columns[2]] = os.path.dirname(devi_file)
                devi_pd = pd.concat([devi_pd, tmp_pd])
    else:
        #step       avg_devi_f       min_devi_f       max_devi_f       avg_devi_e       min_devi_e       max_devi_e
        #   0       0.008682422      0.000574555      0.014187865      0.030333196      0.011702489      0.043806718
        #  10       0.016611307      0.000627126      0.027185410      0.031661788      0.005340899      0.048241921
        for devi_file in model_devi_files:
            devi_force = read_data(devi_file, skiprows=0)
            tmp_pd = pd.DataFrame()
            tmp_pd[EXPLORE_FILE_STRUCTURE.devi_columns[0]] = devi_force[:, 3] # max_devi_f
            tmp_pd[EXPLORE_FILE_STRUCTURE.devi_columns[1]] = devi_force[:, 0]
            tmp_pd[EXPLORE_FILE_STRUCTURE.devi_columns[2]] = os.path.dirname(devi_file)
            devi_pd = pd.concat([devi_pd, tmp_pd])
    devi_pd.reset_index(drop=True, inplace=True)
    devi_pd["config_index"].astype(int)

    return devi_pd, base_force_kpu


def count_info(save_dir, error_pd, accurate_pd, rand_candi, remove_candi, md_run_info:str=None):
    #5. save select info
    total_num = error_pd.shape[0] + accurate_pd.shape[0] + rand_candi.shape[0] + remove_candi.shape[0]
    cand_num = rand_candi.shape[0] + remove_candi.shape[0]
    summary = "Total structures {}    accurate {} rate {:.2f}%    selected {} rate {:.2f}%    error {} rate {:.2f}%\n"\
        .format(total_num, accurate_pd.shape[0], accurate_pd.shape[0]/total_num*100, \
                    cand_num, cand_num/total_num*100, \
                        error_pd.shape[0], error_pd.shape[0]/total_num*100)
    candi_info = ""
    if remove_candi.shape[0] == 0:
        candi_info += "Candidate configurations: {}\n        Select details in file {}\n".format(
            cand_num, EXPLORE_FILE_STRUCTURE.candidate)
    else:
        candi_info += "Candidate configurations: {}, randomly select {}, delete {}\n        Select details in file {}\n        Delete details in file {}.\n".format(
            cand_num, rand_candi.shape[0], remove_candi.shape[0],\
            EXPLORE_FILE_STRUCTURE.candidate, EXPLORE_FILE_STRUCTURE.candidate_delete)
    summary_info = ""
    summary_info += summary
    summary_info += "\nSelect by model deviation force:\n"
    summary_info += "Accurate configurations: {}, details in file {}\n".\
        format(accurate_pd.shape[0], EXPLORE_FILE_STRUCTURE.accurate)
    summary_info += candi_info
    summary_info += "Error configurations: {}, details in file {}\n\n".\
        format(error_pd.shape[0], EXPLORE_FILE_STRUCTURE.failed)
    if md_run_info is not None:
        summary_info += md_run_info
        summary_info += "\n"

        summary += md_run_info 
        summary += "\n"
    if save_dir is not None:
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        accurate_pd.to_csv(os.path.join(save_dir, EXPLORE_FILE_STRUCTURE.accurate))
        rand_candi.to_csv(os.path.join(save_dir, EXPLORE_FILE_STRUCTURE.candidate))
        if remove_candi.shape[0] > 0:
            remove_candi.to_csv(os.path.join(save_dir, EXPLORE_FILE_STRUCTURE.candidate_delete))
        error_pd.to_csv(os.path.join(save_dir, EXPLORE_FILE_STRUCTURE.failed))
        write_to_file(os.path.join(save_dir, EXPLORE_FILE_STRUCTURE.select_summary), summary_info, "w")
        
    return summary_info, summary


def count_mdstop_info(devi_file_list):
    def read_md_last_line(md_log):
        with open(md_log, "rb") as file:
            file.seek(-2, 2)  # 定位到文件末尾前两个字节
            while file.read(1) != b'\n':  # 逐字节向前查找换行符
                file.seek(-2, 1)  # 向前移动两个字节
            last_line = file.readline().decode().strip()  # 读取最后一行并去除换行符和空白字符
        if "Total wall time" in last_line: #md 正常结束
            return True
        else:
            return False
    # for each md model_deviation file get shape
    # for each md md.log get run time 
    # do compare
    right_list = []
    error_list = []
    for devi_file in devi_file_list:
        devi = np.loadtxt(devi_file)
        end_normal = read_md_last_line(os.path.join(os.path.dirname(devi_file), SLURM_OUT.md_out))
        if end_normal and devi.shape[0] > 1:
            right_list.append(os.path.dirname(devi_file))
        else:
            error_list.append(os.path.dirname(devi_file))
    return right_list, error_list

    


import json
import os
from pwact.utils.json_operation import get_parameter, get_required_parameter
from pwact.utils.app_lib.pwmat import is_alive_atomic_energy

class WorkFileStructure(object):
    '''
    description: 
    param {*} self
    param {*} json_dir: The trained models, features, and inference results are collected in this directory
    return {*}
    author: wuxingxing
    '''    
    def __init__(self, json_dir:str, reserve_work_dir:bool, reserve_feature:bool, model_type:str) -> None:
        self.model_type = model_type
        self.json_dir = json_dir
        self.reserve_work_dir = reserve_work_dir
        self.reserve_feature = reserve_feature
        self.movement_name = "MOVEMENT"
        # self.raw_path = []
        self.train_feature_path = []
        self.valid_feature_path = []
        self.test_feature_path = []
        # self.datasets_path = []
        self.model_load_path = ""
        
        self.train_data_path = []
        self.valid_data_path = []
        self.test_data_path  = []
        
        if self.model_type == "NN" or self.model_type == "LINEAR":
            self._set_NN_PWdata_dirs() 
    # def _set_training_path(self, train_raw_path:list, train_feature_path:list, train_dir: str):
    #     self.raw_path = train_raw_path
    #     self.train_feature_path = train_feature_path
    #     self.train_dir = os.path.join(self.json_dir, train_dir)


    # def _set_data_file_paths(self, trainSetDir:str, dRFeatureInputDir:str, dRFeatureOutputDir:str,\
    #                     trainDataPath:str, validDataPath:str):
    #     self.trainSetDir = trainSetDir
    #     self.dRFeatureInputDir = dRFeatureInputDir# it is not used 2024.04.03
    #     self.dRFeatureOutputDir = dRFeatureOutputDir# it is not used 2024.04.03
    #     self.trainDataPath = trainDataPath
    #     self.validDataPath = validDataPath

    def _set_p_matrix_paths(self, p_path, save_p_matrix:bool):
        self.save_p_matrix = save_p_matrix
        self.p_matrix_path = p_path

    def _set_model_paths(self,model_store_dir:str, model_name:str, best_model_path:str):
        self.model_store_dir = model_store_dir
        self.model_name = model_name
        self.best_model_path = best_model_path# it is not used 2024.04.03
        self.model_save_path = os.path.join(model_store_dir, self.model_name)

    def _set_model_load_path(self, model_load_path:str):
        self.model_load_path = model_load_path

    def set_inference_paths(self, json_input:dict, is_nep_txt:bool=False):
        test_dir_name = get_parameter("test_dir_name", json_input, "test_result")
        
        if json_input["model_type"].upper() in ["LINEAR", "NN"]:
            self.test_dir = os.path.join(self.nn_work, test_dir_name)
        else:
            self.test_dir = os.path.join(self.json_dir, test_dir_name)

        if not json_input["model_type"].upper() == "LINEAR":
            if is_nep_txt:
                self.model_load_path = None
            else:
                model_load_path = get_required_parameter("model_load_file", json_input)
                self.model_load_path = os.path.abspath(model_load_path)
                if os.path.exists(self.model_load_path) is False:
                    raise Exception("the model_load_path is not exist: {}, please speccified 'model_load_path' at json file".format(self.model_load_path))
        
        # if "trainDataPath" in json_input.keys():# for test, people could set the 'trainSetDir' to 'valid', so the valid data in train dir could be used for valid
        #     self.trainDataPath = json_input["trainDataPath"]

        '''alive_atomic_energy = is_alive_atomic_energy(datasets_path)
        self._set_alive_atomic_energy(alive_atomic_energy)'''
    

    '''
    description: set feature paths
        the feature path list could from user input.json, \
            the feature path list could auto set after generating feature
    param {*} self
    param {list} feature_path
    return {*}
    author: wuxingxing
    '''
    def set_train_feature_path(self, feature_path:list):
        self.train_feature_path.extend(feature_path)

    def set_valid_feature_path(self, feature_path:list):
        self.valid_feature_path.extend(feature_path)

    def set_test_feature_path(self, feature_path:list):
        self.test_feature_path.extend(feature_path)

    # delete in 2025
    def set_datasets_path(self, datasets_path:list):
        pass
        # self.datasets_path.extend(datasets_path)


    '''
    description: 
    set workdir structrues of dp/NN/linear model when doing initialization
    param {*} self
    param {dict} json_input
    return {*}
    author: wuxingxing
    '''
    def set_model_file_paths(self, json_input:dict):
        model_load_path = get_parameter("model_load_file", json_input, " ")
        if os.path.exists(model_load_path):
            model_load_path = os.path.abspath(model_load_path)
        self._set_model_load_path(model_load_path)
        model_name = ""
        best_model_path = ""
        # if self.recover_train is True and os.path.isfile(model_load_path):
        #     raise Exception("Error! The recover_train and model_load_path are simultaneously specified, please set recover_train to False or remove param model_load_path")
        if self.model_type == "NN":
            model_name = get_parameter("model_name", json_input, "nn_model.ckpt")
        elif self.model_type == "DP":
            model_name = get_parameter("model_name", json_input, "dp_model.ckpt")
        elif self.model_type == "NEP":
            model_name = get_parameter("model_name", json_input, "nep_model.ckpt")
        elif self.model_type == "CHEBY":
            model_name = get_parameter("model_name", json_input, "chey_model.ckpt")
        best_model_path = os.path.join(self.json_dir, "best_model.ckpt")
        forcefield_name = get_parameter("forcefield_name", json_input, "forcefield.ff")
        forcefield_dir = get_parameter("forcefield_dir", json_input, "forcefield")
        # p matix, resume p matrix when recover is not realized
        # p matrix should extract to checkpoint files or a single file.
        # current not realized
        save_p_matrix = get_parameter("save_p_matrix", json_input, False)
        if save_p_matrix is not False:
            Pmatrix_path = os.path.join(self.json_dir, "P.pkl")
            self._set_p_matrix_paths(Pmatrix_path, True)
        else:
            self._set_p_matrix_paths(None, False)

        # common dir 
        model_store_dir = get_parameter("model_store_dir", json_input, "model_record")
        if self.model_type == "NN":
            model_store_dir = os.path.join(self.nn_work, model_store_dir)
            self.forcefield_dir = os.path.join(self.nn_work, forcefield_dir)
            self.forcefield_name = forcefield_name
        else:
            self.forcefield_dir = os.path.join(self.json_dir, forcefield_dir)
            self.forcefield_name = forcefield_name
            model_store_dir = os.path.join(self.json_dir, model_store_dir)
        self._set_model_paths(model_store_dir = model_store_dir, \
                                    model_name = model_name, best_model_path=best_model_path)
        
        # self._set_PWdata_dirs(json_input)

    def set_train_valid_file(self, json_input:dict):
        # set trian movement file path
        self.format = get_parameter("format", json_input, "pwmat/movement").lower() # used in new file and raw_file
        if self.model_type.upper() in ["NN", "LINEAR"]:
            if self.format != "pwmat/movement":
                raise Exception("Error! For NN or Linear model, the input 'format' should be 'pwmat/movement'!")
        train_data = get_parameter("train_data", json_input, [])
        
        for _train_data in train_data:
            if os.path.exists(_train_data) is False:
                raise Exception("Error! train data: {} file not exist!".format(_train_data))
            else:
                self.train_data_path.append(os.path.abspath(_train_data))
        valid_data = get_parameter("valid_data", json_input, [])
        for _valid_data in valid_data:
            if os.path.exists(_valid_data) is False:
                raise Exception("Error! valid data: {} file not exist!".format(_valid_data))
            else:
                self.valid_data_path.append(os.path.abspath(_valid_data))
        test_data = get_parameter("test_data", json_input, [])
        for _test_data in test_data:
            if os.path.exists(_test_data) is False:
                raise Exception("Error! test data: {} file not exist!".format(_test_data))
            else:
                self.test_data_path.append(os.path.abspath(_test_data))

        if self.format == "pwmat/movement": # for nn
            self.alive_atomic_energy = False
            if len(self.train_data_path) > 0:
                alive_atomic_energy = is_alive_atomic_energy(self.train_data_path)
                self.alive_atomic_energy = alive_atomic_energy

            if len(self.valid_data_path) > 0:
                alive_atomic_energy = is_alive_atomic_energy(self.valid_data_path)
                self.alive_atomic_energy = alive_atomic_energy
            
            if len(self.test_data_path) > 0:
                alive_atomic_energy = is_alive_atomic_energy(self.test_data_path)
                self.alive_atomic_energy = alive_atomic_energy

    def set_nn_file(self, json_input:dict):
        self.train_feature_path = []
        self.valid_feature_path = []
        self.test_feature_path  = []
        train_feature_path = get_parameter("train_feature_path", json_input, [])
        for feat_path in train_feature_path:
            if os.path.exists(feat_path) is False:
                raise Exception("Error! train_feature_path: {} file not exist!".format(feat_path))
        self.train_feature_path = [os.path.abspath(_) for _ in train_feature_path]

        valid_feature_path = get_parameter("valid_feature_path", json_input, [])
        for feat_path in valid_feature_path:
            if os.path.exists(feat_path) is False:
                raise Exception("Error! valid_feature_path: {} file not exist!".format(feat_path))
        self.valid_feature_path = [os.path.abspath(_) for _ in valid_feature_path]

        test_feature_path = get_parameter("test_feature_path", json_input, [])
        for feat_path in test_feature_path:
            if os.path.exists(feat_path) is False:
                raise Exception("Error! test_feature_path: {} file not exist!".format(feat_path))
        self.test_feature_path = [os.path.abspath(_) for _ in test_feature_path]
        
    def _set_NN_PWdata_dirs(self):
        # set Pwdata dir file structure, they are used in feature generation
        self.nn_work = os.path.join(os.getcwd(), "work_dir") # the work dir of nn training or test 
        self.trainSetDir = 'PWdata'
        self.dRFeatureInputDir = 'input'# it is not used 2024.04.03
        self.dRFeatureOutputDir = 'output'# it is not used 2024.04.03
        # self.trainDataPath = 'train'
        # self.validDataPath = 'valid'
        # self._set_data_file_paths(trainSetDir, dRFeatureInputDir, dRFeatureOutputDir, trainDataPath, validDataPath)

    def set_nep_native_file_paths(self):
        self.nep_model_file = "nep5.txt"
                                 
    def get_data_file_structure(self):
        file_dict = {}
        file_dict["trainSetDir"] = self.trainSetDir
        file_dict["dRFeatureInputDir"] = self.dRFeatureInputDir
        file_dict["dRFeatureOutputDir"] = self.dRFeatureOutputDir
        # file_dict["trainDataPath"] = self.trainDataPath
        # file_dict["validDataPath"] = self.validDataPath
        return file_dict

    def to_dict(self):
        dicts = {}
        dicts["format"] = self.format     
        if self.model_load_path is not None and os.path.exists(self.model_load_path):
            dicts["model_load_file"] = self.model_load_path
        if len(self.train_data_path) > 0:
            dicts["train_data"] = self.train_data_path
        if len(self.valid_data_path) > 0:
            dicts["valid_data"] = self.valid_data_path
        if len(self.test_data_path) > 0:
            dicts["test_data"] = self.test_data_path
        return dicts



import os
from pwact.utils.constant import ELEMENTTABLE, DFT_STYLE, ELEMENTTABLE_2, CP2K, PWDATA
from pwact.utils.app_lib.cp2k import make_cp2k_xyz
from pwact.utils.file_operation import write_to_file
from pwdata.config import Config
from pwdata import perturb_structure, make_supercell, scale_cell
'''
description: 
    return atom type names and atom type numbers of config file
param {str} config_path
param {str} format
return {*}
author: wuxingxing
'''
def get_atom_type(config_path, format:str=None):
    if isinstance(config_path, str):
        image = Config(format=format, data_path=config_path, atom_names=None).images[0]
    else:
        image = config_path
    atomic_number_list = []
    atomic_name_list = []
    for atom in image.atom_type:
        atomic_name_list.append(ELEMENTTABLE_2[atom])
        atomic_number_list.append(atom)
    return atomic_name_list, atomic_number_list

def load_config(config, format, atom_names=None):
    config = Config(format=format, data_path=config, atom_names=atom_names)
    return config

'''
description: 
    the input config could be vasp/pwmat/cp2k format
    for cp2k, the save_name will be 'coord.xyz'
return {*}
author: wuxingxing
'''
def save_config(config, input_format:str = None, wrap = False, direct = True, sort = True, \
        save_format:str=None, save_path:str=None, save_name:str=None, atom_names: list[str] = None):
    if isinstance(config, str):
        config = Config(format=input_format, data_path=config, atom_names=atom_names).images[0]
    if save_format == PWDATA.cp2k_scf:
        # make coord.xyz used by cp2k for every task 
        config = config._set_cartesian() if config.cartesian is False else config._set_cartesian()
        # potential = {"Si":"GTH-PBE"},
        # basis_set = {"Si":"DZVP-MOLOPT-SR-GTH-q4"}
        atom_types_image = []
        for atom in config.atom_types_image:
            atom_types_image.append(ELEMENTTABLE_2[atom])
        coord_xyz = make_cp2k_xyz(
            atom_types = atom_types_image,
            coord_list = config.position
        )
        write_to_file(os.path.join(save_path, save_name), coord_xyz, 'w')

        lattice = [element for sublist in config.lattice for element in sublist]
        lattice_line = ",".join(str(_) for _ in lattice)
        write_to_file(os.path.join(save_path, CP2K.cell_txt), lattice_line, 'w')

    else:
        config.to(data_path  =save_path, 
                data_name    =save_name, 
                format       =save_format, 
                direct       =direct, 
                sort         =sort, 
                wrap         =wrap
                )
    
    return os.path.join(save_path, save_name)

def read_cp2k_xyz(config_file:str):
    with open(config_file, 'r') as rf:
            config_contents = rf.readlines()
    atom_names = []
    atom_type_name = []
    coord = []
    for line in config_contents:
        if line.strip() == 0:
            continue
        elements = line.split()
        atom_names.append(elements[0])
        if elements[0] not in atom_type_name:
            atom_type_name.append(elements[0])
        coord.appnd([float(elements[1]), float(elements[2]), float(elements[3])])
    return atom_type_name, atom_names, coord

def do_super_cell(config_file, input_format:str=None, supercell_matrix:list[int]=None, pbc:list[int]=[1, 1, 1], direct = True, sort = True, \
                    save_format:str=None, save_path:str=None, save_name:str=None):
    config = Config(format=input_format, data_path=config_file, atom_names=None)
    # Make a supercell     
    supercell = make_supercell(config, supercell_matrix, pbc)
    # Write out the structure
    supercell.to(data_path = save_path,
                data_name  = save_name,
                format     = save_format,
                direct = direct,
                sort = sort)
    return os.path.join(save_path, save_name)

def do_scale(config, input_format:str=None, scale_factor:float=None, 
            direct:bool=True, sort:bool=True, save_format:str=None, save_path:str=None, save_name:str=None):
    config = Config(format=input_format, data_path=config)
    scaled_struct = scale_cell(config, scale_factor)
    scaled_struct.to(data_path = save_path,
                    data_name  = save_name,
                    format     = save_format,
                    direct = direct,
                    sort = sort)
     
    return os.path.join(save_path, save_name)

def do_pertub(config, input_format:str=None, pert_num:int=None, cell_pert_fraction:float=None, atom_pert_distance:float=None, \
        direct:bool=True, sort:bool=True, save_format:str=None, save_path:str=None, save_name:str=None):
    config = Config(format=input_format, data_path=config)

    if not os.path.exists(save_path):
        os.makedirs(save_path)
        
    perturbed_structs = perturb_structure(
            image_data = config,
            pert_num = pert_num,
            cell_pert_fraction = cell_pert_fraction,
            atom_pert_distance = atom_pert_distance)

    for tmp_perturbed_idx, tmp_pertubed_struct in enumerate(perturbed_structs):
        tmp_pertubed_struct.to(data_path  = save_path,
                                data_name = "{}_{}".format(tmp_perturbed_idx, save_name),
                                format    = save_format,
                                direct = direct,
                                sort = sort)

        print("pertub {} done!".format(os.path.join(save_path, "{}_{}".format(tmp_perturbed_idx, save_name))))

'''
description: 
    save the inputfiles to pwmlff/npy format data
return {*}
author: wuxingxing
'''
def extract_pwdata(input_data_list:list[str], 
                intput_data_format:str="pwmat/movement", 
                save_data_path:str="./",
                save_data_name="PWdata", 
                save_data_format="extxyz",
                data_shuffle:bool=False,
                interval:int=1
                ):
    # if data_format == DFT_STYLE.cp2k:
    #     raise Exception("not relized cp2k pwdata convert")

    if not os.path.isabs(save_data_path):
        # data_name = datasets_path
        save_data_path = os.path.join(os.getcwd(), save_data_path)
    image_data = None
    for dir in input_data_list:
        try:
            if image_data is not None:
                tmp_config = Config(format=intput_data_format, data_path=dir)
                # if not isinstance(tmp_config, list):
                #     tmp_config = [tmp_config]
                image_data.images.extend(tmp_config.images)
            else:
                image_data = Config(format=intput_data_format, data_path=dir)
                
                if not isinstance(image_data.images, list):
                    image_data.images = [image_data.images]
        except Exception as e:
            print("Warning! Read the traj file error {}".format(dir))
            pass
            # if not isinstance(image_data, list):
            #     image_data = [image_data]
    if interval > 1:
        tmp = []
        for i in range(0, len(image_data.images)):
            if i % interval == 0:
                tmp.append(image_data.images[i])
        image_data.images = tmp

    image_data.to(
                data_path  =save_data_path,
                data_name  =save_data_name,
                format     =save_data_format,
                random=data_shuffle
                )
    
if __name__ == "__main__":
    # in_config = "/data/home/wuxingxing/datas/al_dir/si_exp/init_bulk/collection/init_config_1/0.9_scale_pertub/0_pertub.config"
    # save_path = "/data/home/wuxingxing/datas/al_dir/si_exp/init_bulk/collection/init_config_1/0.9_scale_pertub"
    # image = Config.read(format="pwmat", data_path=in_config)
    # save_config(image, wrap=False, direct=True, sort=True,\
    #     save_format="vasp", save_path=save_path, save_name="temp_poscar")
    # save_config(image, wrap=False, direct=True, sort=True,\
    #     save_format="vasp", save_path=save_path, save_name="temp_poscar")
    
    # do_super_cell(config=image,
    #         supercell_matrix=[[2,0,0], [0,2,0], [0,0,2]], 
    #         pbc=[1, 1, 1], 
    #         direct = True, 
    #         sort = True, \
    #         save_format="pwmat", save_path=save_path, save_name="temp_super_atom.config")
    
    # do_super_cell(config="/data/home/wuxingxing/datas/al_dir/si_5_pwmat/sys_config/structures/44_POSCAR",
    #     input_format="vasp",
    #     supercell_matrix=[[2,0,0], [0,2,0], [0,0,2]], 
    #     pbc=[1, 1, 1], 
    #     direct = True, 
    #     sort = True, \
    #     save_format="pwmat", save_path=save_path, save_name="temp_super_atom_from_poscar.config")
    
    
    # do_super_cell(config="/data/home/wuxingxing/datas/al_dir/si_5_pwmat/sys_config/structures/44_POSCAR",
    #     input_format="vasp",
    #     supercell_matrix=[[2,0,0], [0,2,0], [0,0,2]], 
    #     pbc=[1, 1, 1], 
    #     direct = True, 
    #     sort = True, \
    #     save_format="vasp", save_path=save_path, save_name="temp_super_poscar")

    # do_scale(config=image,
    #         scale_factor=0.99, 
    #         direct=True,
    #         sort=True, 
    #         save_format="pwmat", save_path=save_path, save_name="temp_scale_atom.config")

    # do_scale(config=image,
    #         scale_factor=0.99, 
    #         direct=True,
    #         sort=True, 
    #         save_format="vasp", save_path=save_path, save_name="temp_scale_poscar")
    
    # save_path = "/data/home/wuxingxing/datas/al_dir/si_5_pwmat/sys_config/structures/temp_pwmat_pertub"
    # do_pertub(config=image, 
    #     pert_num=50, 
    #     cell_pert_fraction=0.01, 
    #     atom_pert_distance=0.04, 
    #     direct=True,
    #     sort=True, 
    #     save_format="pwmat", save_path=save_path, save_name="pertub.config")
    
    # save_path = "/data/home/wuxingxing/datas/al_dir/si_5_pwmat/sys_config/structures/temp_vasp_pertub"
    # do_pertub(config=image, 
    #     pert_num=50, 
    #     cell_pert_fraction=0.01, 
    #     atom_pert_distance=0.04, 
    #     direct=True,
    #     sort=True, 
    #     save_format="vasp", save_path=save_path, save_name="pertub.poscar")

    # convert trajs to config for model_deviation calculate
    # import glob
    # traj_dir = "/data/home/wuxingxing/datas/al_dir/si_5_pwmat/iter.0000/temp_run_iter_work/explore/md/md.000.sys.000/md.000.sys.000.t.000/traj"
    # save_dir = "/data/home/wuxingxing/datas/al_dir/si_5_pwmat/iter.0000/temp_run_iter_work/explore/md/md.000.sys.000/md.000.sys.000.t.000/traj2config"
    # if not os.path.exists(save_dir):
    #     os.makedirs(save_dir)
    # trajs = glob.glob(os.path.join(traj_dir, "*.lammpstrj"))
    # trajs = sorted(trajs)
    # for traj in trajs:
    #     save_name = "{}.config".format(os.path.basename(traj).split('.')[0])
    #     save_config(traj,  input_format="dump", wrap=False, direct=True, sort=True,\
    #     save_format="pwmat", save_path=save_dir, save_name=save_name)
    #     print("{} to {} done!".format(traj, save_name))
    import glob
    data_dir1 = glob.glob(os.path.join("/data/home/wuxingxing/datas/al_dir/HfO2/dftb/init_bulk_hfo2/temp_init_bulk_work/scf/init_config_*")) #
    data_dir2 = glob.glob(os.path.join("/data/home/wuxingxing/datas/al_dir/HfO2/dftb/init_bulk_hfo2_600k/temp_init_bulk_work/scf/init_config_*")) 
    data_dir1.extend(data_dir2)
    select_list = ["0-scf", "200-scf", "400-scf", "600-scf", "800-scf", "1000-scf"]
    data_list = []
    for dir in data_dir1:
        _outcars = glob.glob(os.path.join(dir, "*/*/*/OUTCAR"))
        for outcar in _outcars:
            if os.path.basename(os.path.dirname(outcar)) in select_list:
                data_list.append(outcar)

    datasets_path = "/data/home/wuxingxing/datas/al_dir/HfO2/dftb/init_data_200"
    extract_pwdata(input_data_list=data_list, 
            intput_data_format="vasp/outcar", 
            save_data_path=datasets_path
            )


#!/usr/bin/env python
import os
import signal
import glob
import sys
import json
import argparse
from pwact.utils.constant import TEMP_STRUCTURE, UNCERTAINTY, AL_WORK, AL_STRUCTURE, LABEL_FILE_STRUCTURE, EXPLORE_FILE_STRUCTURE, DFT_STYLE
from pwact.utils.format_input_output import make_iter_name
from pwact.utils.file_operation import write_to_file, del_file_list, search_files, del_dir, copy_dir, copy_file
from pwact.utils.json_operation import convert_keys_to_lowercase

from pwact.active_learning.user_input.resource import Resource
from pwact.active_learning.user_input.iter_input import InputParam
from pwact.active_learning.user_input.init_bulk_input import InitBulkParam
from pwact.active_learning.train.train_model import ModelTrian
from pwact.active_learning.train.dp_kpu import ModelKPU
from pwact.active_learning.explore.run_model_md import Explore
from pwact.active_learning.label.labeling import Labeling
from pwact.active_learning.user_input.cmd_infos import cmd_infos

from pwact.active_learning.init_bulk.init_bulk_run import init_bulk_run, scancel_jobs as init_scancel_jobs
from pwact.active_learning.environment import check_envs

from pwact.data_format.configop import extract_pwdata
from pwact.active_learning.explore.select_image import select_image, print_select_image
from pwact.utils.process_tool import kill_process, get_pid
def run_iter():
    system_json = json.load(open(sys.argv[2]))
    if "work_dir" in system_json.keys():
        os.chdir(system_json["work_dir"])
    pid = get_pid()
    with open("./PID", 'w') as wf:
        wf.write(pid)

    system_info = convert_keys_to_lowercase(system_json)
    machine_json = json.load(open(sys.argv[3]))
    machine_info = convert_keys_to_lowercase(machine_json)

    input_param = InputParam(system_info)
    resource = Resource(machine_info, dft_style=input_param.dft_style)
    os.chdir(input_param.root_dir)
    print("The work dir change to {}".format(os.getcwd()))
    record = os.path.abspath(input_param.record_file)
    iter_rec = [0, -1]
    if os.path.isfile(record):
        with open (record) as frec :
            for line in frec :
                if line == '\n':
                    continue
                iter_rec = [int(x) for x in line.split()]
        print ("continue from iter %03d task %02d" % (iter_rec[0], iter_rec[1]))

    ii = -1
    numb_task = 3
    max_tasks = input_param.explore.md_job_num
    while ii < max_tasks:#control by config.json
        ii += 1
        iter_name=make_iter_name(ii)
        for jj in range (numb_task) :
            if ii * max_tasks*10000 + jj <= iter_rec[0] * max_tasks*10000 + iter_rec[1] : 
                continue
            print("current iter is {}".format(iter_name))
            task_name="task %02d"%jj
            print("{} - {}".format(iter_name, task_name))
            if  jj == 0 and ii <= max_tasks: # the last iter, only need to train the model with all datas
                print ("training start: iter {} - task {}".format(ii, jj))
                do_training_work(iter_name, resource, input_param)
                write_to_file(record, "\n{} {}".format(ii, jj), "a") #record_iter
            elif jj == 1 and ii < max_tasks:
                print ("exploring start: iter {} - task {}".format(ii, jj))
                do_exploring_work(iter_name, resource, input_param)
                write_to_file(record, "\n{} {}".format(ii, jj), "a") #record_iter
            elif jj == 2 and ii < max_tasks:
                print ("run_fp: iter {} - task {}".format(ii, jj))
                run_fp(iter_name, resource, input_param)
                write_to_file(record, "\n{} {}".format(ii, jj), "a") #record_iter
            # write_to_file(record, "\n{} {}".format(ii, jj), "a")
            if jj == 2 and not input_param.reserve_work: # delete temp_work_dir under current iteration after the labeling done
                del_file_list([os.path.join(input_param.root_dir, iter_name, TEMP_STRUCTURE.tmp_run_iter_dir)])

    print("Active learning done! \nYou could use cmd 'pwact gather_pwdata' to collect all datas sampled from iterations.")

def run_fp(itername:str, resource : Resource, input_param: InputParam):
    lab = Labeling(itername, resource, input_param)
    #1. if the label work done before, back up and do new work
    lab.back_label()
    #2. make scf work
    if input_param.dft_style == DFT_STYLE.bigmodel:
        lab.make_bigmodel_work()
        lab.do_bigmodel_jobs()
        lab.do_post_bigmodel()
    else:
        lab.make_scf_work()
        #3. do scf work
        lab.do_scf_jobs()
        #4. collect scf configs outcar or movement, then to pwdata format
        lab.do_post_labeling()
    
def do_training_work(itername:str, resource : Resource, input_param: InputParam):
    mtrain = ModelTrian(itername, resource, input_param)
    # 1. if the train work done before, backup the train dir and retrain
    mtrain.back_train()
    # 2. create train work dirs
    mtrain.make_train_work()
    # 3. run training job
    mtrain.do_train_job()
    # 4. do post process after training
    mtrain.post_process_train()
    print("{} done !".format("train_model"))

def do_exploring_work(itername:str, resource : Resource, input_param: InputParam):
    md = Explore(itername, resource, input_param)
    # 1. if the explore work done before, back up explore dir do new explore work
    md.back_explore()
    # 2. make md work files
    md.make_md_work()
    # 3. do md job
    md.do_md_jobs()
    # 4. select images
    if input_param.strategy.uncertainty.upper() == UNCERTAINTY.committee.upper():
        summary = select_image(
                md_dir=md.md_dir, 
                save_dir=md.select_dir,
                md_job=md.md_job,
                devi_name=EXPLORE_FILE_STRUCTURE.get_devi_name(UNCERTAINTY.committee),
                lower=input_param.strategy.lower_model_devi_f,  
                higer=input_param.strategy.upper_model_devi_f
        )
        # summary = md.select_image_by_committee()
        # committee: read model deviation file under md file
    elif input_param.strategy.uncertainty.upper() == UNCERTAINTY.kpu.upper():
        summary = uncertainty_analyse_kpu(itername, resource, input_param)
    summary = "{}  {}\n".format(itername, summary)
    write_to_file(os.path.join(input_param.root_dir, EXPLORE_FILE_STRUCTURE.iter_select_file), summary, mode='a')

    if input_param.strategy.direct:
        md.make_drct_work()
        md.do_drct_jobs()
        md.post_drct()
    print("config selection done!")

    # 5. do post process 
    md.post_process_md()
    print("exploring done!")

def uncertainty_analyse_kpu(itername:str, resource : Resource, input_param: InputParam):
    mkpu = ModelKPU(itername, resource, input_param)
    # 1. make kpu work dirs
    mkpu.make_kpu_work()
    # 2. do kpu job
    mkpu.do_kpu_jobs()
    # 3. post process after kpu calculate: select images
    summary = mkpu.post_process_kpu()
    return summary

def init_bulk():
    system_json = json.load(open(sys.argv[2]))
    system_info = convert_keys_to_lowercase(system_json)
    if "work_dir" in system_json.keys():
        os.chdir(system_json["work_dir"])
    pid = get_pid()
    with open("./PID", 'w') as wf:
        wf.write(pid)

    machine_info = convert_keys_to_lowercase(json.load(open(sys.argv[3])))
    input_param = InitBulkParam(system_info)
    resource = Resource(machine_info, job_type=AL_WORK.init_bulk, dft_style=input_param.dft_style, scf_style=input_param.scf_style)

    os.chdir(input_param.root_dir)
    
    print("The work dir change to {}".format(os.getcwd()))
    init_bulk_run(resource, input_param)
    print("Init Bulk Work Done!")

def gather_pwmata(input_cmds):
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input_dir', help="specify the dir above the iterations, the default dir is current dir './'\nthe result could be found in './final_pwdata'", type=str, default='./')
    args = parser.parse_args(input_cmds)
    if not os.path.exists(args.input_dir):
        raise Exception("Error! The input dir {} not exists!".format(args.input_dir))
    pwdata_lists = search_files(args.input_dir, "iter*/{}/{}/*".format(AL_STRUCTURE.labeling,  LABEL_FILE_STRUCTURE.result))
    pwdata_lists = sorted(pwdata_lists)
    save_dir = "./final_pwdata"
    res_data_list = []
    for pwdata in pwdata_lists: # /path/iter.0001/label/result/*
        data_name = os.path.basename(pwdata) 
        iter_name = os.path.basename(os.path.dirname(os.path.dirname(os.path.dirname(pwdata)))) #iter.0001
        target_dir = os.path.join(save_dir, iter_name, data_name) #./final_pwdata/iter.0001/train.xyz or ./final_pwdata/iter.0001/pwdata/...
        if os.path.isfile(pwdata) and data_name == "train.xyz":
            copy_file(pwdata, target_dir)
            res_data_list.append(os.path.realpath(target_dir))
        else:
            copy_dir(pwdata, target_dir)
            for root, dirs, files in os.walk(pwdata):
                    if 'energies.npy' in files:
                        res_data_list.append(root)
        
    result_lines = ["\"{}\",".format(_) for _ in res_data_list]
    result_lines = "\n".join(result_lines)
    # result_lines = result_lines[:-1] # Filter the last ','
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
    result_save_path = os.path.join(save_dir, "final_pwdata_list.txt")
    write_to_file(result_save_path, result_lines, mode='w')
    print("All datas in iterations are:\n")
    print(result_lines)
    print("more details could be found in dir {}\n".format(save_dir))

def print_init_json_template():
    pass

def init_surface():
    pass

'''
description: 
输出主动学习流程的模板json文件
return {*}
author: wuxingxing
'''
def print_run_json_template():
    pass

'''
description: 
    增加一些小工具
        1. 整理出所有的movement到一个目录？
return {*}
author: wuxingxing
'''
def common_tool():
    pass

def environment_check():
    check_envs()

def run_scancel_jobs(iter:int, step:int):
    # train
    itername = make_iter_name(iter)
    if step == 0:
        ModelTrian.kill_job(os.getcwd(), itername)
    # explore
    elif step == 1:
        Explore.kill_job(os.getcwd(), itername)
    # labeling
    elif step ==2:
        Labeling.kill_job(os.getcwd(), itername)

def kill_job():
    # system_json = json.load(open(sys.argv[3]))
    # if "work_dir" in system_json.keys():
    #     os.chdir(system_json["work_dir"])
    try:
        with open("./PID", 'r') as rf:
            pid_str_info = rf.readline().split()
            pid = pid_str_info[1]
            jobid = pid_str_info[3] if "job" in pid_str_info else None
    except :
        raise Exception("Error parsing PID file !")
    kill_process(pid, jobid)
    if sys.argv[2].lower() == "init_bulk":
        # search all jobs
        init_scancel_jobs(os.getcwd())
    elif sys.argv[2].lower() == "run":
        iter_rec = [0, -1]
        if os.path.isfile("./al.record"):
            with open ("./al.record") as frec :
                for line in frec :
                    if line == '\n':
                        continue
                    iter_rec = [int(x) for x in line.split()]
                if iter_rec[1] == 2:
                    iter_rec = [iter_rec[0] + 1, 0]
                else:
                    iter_rec = [iter_rec[0], iter_rec[1] + 1]
            print ("The iter %03d task %02d" % (iter_rec[0], iter_rec[1]))
        else:
            iter_rec = [0, 0]
        run_scancel_jobs(iter_rec[0], iter_rec[1])
    else:
        error_log = "Error parsing command !"
        error_log += "The command of kill supported as 'pwact kill init_bulk' or 'pwact kill run <record.file>'\n"
        raise Exception(error_log)

        # for init_bulk relax or aimd jobs
        
    # for run iters jobs

def filter_test(input_cmds):
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--md_dir', help="specify the md dir such as 'iter.0000/temp_run_iter_work/explore/md' or iter.0000/explore/md", type=str, required=True)
    parser.add_argument('-l', '--lower', help="specify lower limit value", type=float, required=True)
    parser.add_argument('-u', '--upper', help="specify upper limit value", type=float, required=True)
    parser.add_argument('-s', '--save', action='store_true', help="if '-s' is set, save the detailed information of the selected configs to CSV files")
    
    args = parser.parse_args(input_cmds)
    if not os.path.exists(args.md_dir):
        raise Exception("ERROR! The input md_dir {} not found!".format(args.md_dir))

    save_dir = os.path.join(os.getcwd(), "filter_test_result") if args.save else None
    summary = print_select_image(
                md_dir=args.md_dir, 
                save_dir=save_dir,
                devi_name=EXPLORE_FILE_STRUCTURE.get_devi_name(UNCERTAINTY.committee),
                lower=args.lower,  
                higer=args.upper
        )

def main():
    environment_check()
    if len(sys.argv) == 1 or "-h".upper() == sys.argv[1].upper() or \
        "help".upper() == sys.argv[1].upper() or "-help".upper() == sys.argv[1].upper() or "--help".upper() == sys.argv[1].upper():
        cmd_infos()

    elif "init_bulk".upper() == sys.argv[1].upper():
        if len(sys.argv) == 2 or "-h".upper() == sys.argv[2].upper() or \
        "help".upper() == sys.argv[2].upper() or "-help".upper() == sys.argv[2].upper() or "--help".upper() == sys.argv[2].upper():
            cmd_infos("init_bulk")
        else:
            init_bulk()
    elif "draw".upper() == sys.argv[1].upper():
        from pwact.active_learning.draw.draw_pictures import draw_pictures
        draw_pictures(sys.argv[2:])

    elif "int_surface".upper() == sys.argv[1].upper():
        init_surface()

    elif "init_json".upper() == sys.argv[1].upper():
        print_init_json_template()

    elif "run_json".upper() == sys.argv[1].upper():
        print_run_json_template()
    
    elif "gather_pwdata".upper() == sys.argv[1].upper():
        gather_pwmata(sys.argv[2:])

    elif "to_pwdata".upper() == sys.argv[1].upper():#these function may use pwdata command
        print("\n\nWarning! This method has been abandoned, new conversion methods refer to the pwdata documentation http://doc.lonxun.com/PWMLFF/Appendix-2/\n\n")
        # to_pwdata(sys.argv[2:])
 
    elif "run".upper() == sys.argv[1].upper():
        if len(sys.argv) == 2 or "-h".upper() == sys.argv[2].upper() or \
        "help".upper() == sys.argv[2].upper() or "-help".upper() == sys.argv[2].upper() or "--help".upper() == sys.argv[2].upper():
            cmd_infos("run")
        else:
            run_iter()
    
    elif "kill".upper() == sys.argv[1].upper():
        if len(sys.argv) == 2 or "-h".upper() == sys.argv[2].upper() or \
        "help".upper() == sys.argv[2].upper() or "-help".upper() == sys.argv[2].upper() or "--help".upper() == sys.argv[2].upper():
            cmd_infos("kill")
        else:
            kill_job()
    
    elif "filter_test".upper() == sys.argv[1].upper() or "filter".upper() == sys.argv[1].upper():
        if len(sys.argv) == 2 or "-h".upper() == sys.argv[2].upper() or \
            "help".upper() == sys.argv[2].upper() or "-help".upper() == sys.argv[2].upper() or "--help".upper() == sys.argv[2].upper():
            cmd_infos("filter")
        else:
            filter_test(sys.argv[2:])

    else:
        print("ERROR! The input cmd {} can not be recognized, please check.".format(sys.argv[1]))
        print("\n\n\nYou can enter the following command.\n\n\n")
        cmd_infos()

if __name__ == "__main__":
    main()
